import sys

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

import threading
from typing import Callable, Dict, Generic, Mapping, Optional, Type, TypeVar, Union

from .types import Options

A = TypeVar("A", covariant=True)
R = TypeVar("R", covariant=True, bound="Request")


lock = threading.Lock()


Handler = Callable[[R], A]


class Request(Generic[A]):
    """Base class for all requests.

    Requests are objects that are passed to the runtime to be handled by a
    handler. The handler is responsible for processing the request and returning
    a result. Requests are used to represent side effects using data, and allow
    the runtime to execute the side effects in a controlled manner.
    """

    options: Options

    def run(self) -> A:
        """Runs the request using the current runtime, returning the result.

        This method should not be overridden.
        """
        return current_runtime().run(self)

    @classmethod
    def handle(cls, handler: Handler[Self, A]) -> Handler[Self, A]:
        """Decorator to register a default handler for a request type.

        This method should not be overridden.

        Arguments
        ----------
        handler : Handler[Self, A]
            The handler to register for the request type.

        Returns
        -------
        Handler[Self, A]
            The handler that was registered.


        Example Usage
        -------------
        >>> class PrintRequest(Request[None]):
        ...     msg: str
        ...     def __init__(self, msg: str, options: Options) -> None:
        ...         self.msg = msg
        ...         self.options = options
        >>>
        >>> @MyRequest.handle
        ... def print_handler(request: MyRequest) -> str:
        ...     print(request.msg)
        >>>
        >>> PrintRequest("Hello, world!", {}).run()
        Hello, world!
        """
        handle_by_default(cls, handler)
        return handler


_DEFAULT_HANDLERS: Dict[Type[Request], Handler] = {}


class Runtime:
    """A class that manages the execution of requests via handlers.

    Runtimes are used to execute side effects in a controlled manner. They
    maintain a mapping of request types to handlers, and execute the appropriate
    handler for each request.

    Can be used as a context manager, which sets the current runtime for the
    duration of the context block for the current thread.

    Arguments
    ----------
    handlers : Optional[Mapping[Type[Request], Handler]]
        A mapping of request types to handlers. If provided, these handlers will
        be used in place of the default handlers.
    """

    handlers: Mapping[Type[Request], Handler]
    previous: Optional["Runtime"]

    def __init__(
        self, handlers: Optional[Mapping[Type[Request], Handler]] = None
    ) -> None:
        self.handlers = {**_DEFAULT_HANDLERS, **(handlers or {})}
        self.previous = None

    def handle(
        self,
        request: Union[Type[R], Mapping[Type[Request], Handler]],
        handler: Optional[Handler[R, A]] = None,
    ) -> "Runtime":
        """Creates a new runtime with the provided request type and handler.

        This method returns a new runtime with the provided request type and
        handler registered. The new runtime will inherit the handlers from the
        current runtime, and will override any handlers for the provided request
        type.

        Arguments
        ----------
        request : Union[Type[R], Mapping[Type[Request], Handler]]
            The request type to handle, or a mapping of request types to handlers.
        handler : Optional[Handler[R, A]]
            The handler to register for the request type.

        Returns
        -------
        Runtime
            A new runtime with the provided request type and handler registered.

        Raises
        ------
        TypeError
            If the request is not a request type or a mapping of request types to
            handlers.
        """
        if isinstance(request, Mapping):
            return Runtime({**self.handlers, **request})
        elif isinstance(request, type) and callable(handler):
            return Runtime({**self.handlers, request: handler})  # type: ignore

        raise TypeError(
            "Runtime.handle() requires either a request type and handler, "
            "or a mapping of request types to handlers."
        )

    def run(self, request: Request[A]) -> A:
        """Runs the provided request using the registered handler.

        This is not meant to be called directly. Instead, use the `run()` method
        on the request object.

        Arguments
        ----------
        request : Request[A]
            The request to run.

        Returns
        -------
        A
            The result of running the request with the correct handler.
        """

        try:
            handler = self.handlers[type(request)]
        except KeyError as e:
            raise TypeError(
                f"No handler for request type {type(request).__qualname__}"
            ) from e

        return handler(request)

    def __enter__(self):
        with lock:
            self.previous = _RUNTIMES.get(threading.current_thread())
            _RUNTIMES[threading.current_thread()] = self
            return self

    def __exit__(self, exc_type, exc_value, traceback):
        with lock:
            _RUNTIMES[threading.current_thread()] = self.previous
            self.previous = None


_RUNTIMES: Dict[threading.Thread, Runtime] = {}


def current_runtime() -> Runtime:
    """Returns the current runtime for the current thread."""
    with lock:
        return _RUNTIMES.setdefault(threading.current_thread(), Runtime())


def handle(
    request: Union[Type[R], Mapping[Type[Request], Handler]],
    handler: Optional[Handler[R, A]] = None,
) -> Runtime:
    """Alias for current_runtime().handle(request, handler).

    This is useful when using the runtime as a context manager.

    Arguments
    ----------
    request : Union[Type[R], Mapping[Type[Request], Handler]]
        The request type to handle, or a mapping of request types to handlers.
    handler : Optional[Handler[R, A]]
        The handler to register for the request type.

    Returns
    -------
    Runtime
        A new runtime with the provided request type and handler registered


    Example Usage
    -------------
    >>> with handle(MyRequest, my_handler):
    ...     MyRequest("Hello, world!", {}).run()
    """
    return current_runtime().handle(request, handler)


def handle_by_default(request: Type[R], handler: Handler[R, A]) -> None:
    """Registers a default handler for a request type.

    Arguments
    ----------
    request : Type[R]
        The request type to handle.
    handler : Handler[R, A]
        The handler to register for the request type.
    """
    with lock:
        _DEFAULT_HANDLERS[request] = handler


def inherit(parent: threading.Thread) -> None:
    """Inherit the runtime from a parent thread.

    Arguments
    ----------
    parent : threading.Thread
        The parent thread to inherit the runtime from.
    """
    with lock:
        _RUNTIMES[threading.current_thread()] = _RUNTIMES.get(parent, Runtime())



import functools
import operator
from types import new_class
from typing import Any, Dict, Optional, Set, Type, TypeVar

from confectioner.templating import set_dotted_key

from .types import Evaluatable, Options, Value

A = TypeVar("A")


class _DatasetClassMeta(type, Evaluatable[A]):
    """Metaclass for DatasetClass objects."""

    def __init__(cls, *args, **kwargs):
        annotations = functools.reduce(
            operator.or_,
            (
                getattr(base, "__annotations__", {})
                for base in reversed(cls.__bases__)
                if base is not _DatasetClassMixin
            ),
        )

        for key in annotations.keys():
            try:
                val = getattr(cls, key)
            except AttributeError:
                raise ValueError(
                    f"Annotation {key} on class {cls.__name__} has "
                    f"no default value."
                )

            if not isinstance(val, Evaluatable):
                setattr(cls, key, Value(val))

        super().__init__(*args, **kwargs)

    def evaluate(cls, options: Options) -> A:
        return cls(options)

    def validate(cls, options: Options) -> None:
        for key in dir(cls):
            dependency = getattr(cls, key, None)
            if isinstance(dependency, Evaluatable) and not key.startswith("__"):
                dependency.validate(options)

    def keys(cls, options: Options) -> Set[str]:
        return {
            key
            for name in dir(cls)
            if (
                isinstance(getattr(cls, name), Evaluatable)
                and not name.startswith("__")
            )
            for key in getattr(cls, name).keys(options)
        }

    def explain(cls, options: Optional[Options] = None) -> Set[str]:
        return {
            key
            for name in dir(cls)
            if (
                isinstance(getattr(cls, name), Evaluatable)
                and not name.startswith("__")
            )
            for key in getattr(cls, name).explain(options)
        }

    def __subclasses__(cls=None):
        return []

    @property
    def result(cls) -> A:
        return cls  # type: ignore

    def __repr__(self):
        return f"<DatasetClass {self.__name__}>"


class _DatasetClassMixin:
    """Mixin for DatasetClass objects."""

    _repr_options: Dict[str, Any]

    def __init__(self, options: Optional[Options] = None):
        options = options or {}

        key: str
        val: Any
        for key in dir(self.__class__):
            val = getattr(self, key)
            if isinstance(val, Evaluatable) and not key.startswith("__"):
                setattr(self, key, val.evaluate(options))

        self._repr_options = {}
        for key in sorted(self.__class__.keys(options)):  # type: ignore [attr-defined]
            value = options.get(key)
            set_dotted_key(key, value, self._repr_options)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._repr_options!r})"

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self._repr_options == other._repr_options
        )


def datasetclass(c: Type[A]) -> _DatasetClassMeta[A]:
    """Create a new DatasetClass from a class definition.

    DatasetClasses are classes that when instatiated, evaluate their
    members using the options dict provided to the constructor. This
    allows for the definition of complex data structures that can be
    evaluated at runtime.

    Any members of the class that are not Evaluatables are wrapped in
    :class:`labrea.Value` instances.

    Example Usage
    -------------
    >>> from labrea import dataset, datasetclass, Option
    >>> @dataset
    ... def my_dataset(a: str = Option('A')) -> str:
    ...     return a
    >>>
    >>> @datasetclass
    ... class MyDatasetClass:
    ...     a: str = my_dataset
    ...     b: int = Option('B')
    ...     c: bool = True
    ...
    >>> inst = MyDatasetClass({'A': 'Hello World!', 'B': 1})
    >>> print(inst.a, inst.b, inst.c)
    Hello World! 1 True
    """
    dataset_class = new_class(
        c.__name__, (c, _DatasetClassMixin), kwds={"metaclass": _DatasetClassMeta}
    )

    functools.update_wrapper(dataset_class, c, updated=())

    return dataset_class  # type: ignore



from labrea.exceptions import KeyNotFoundError
from labrea.option import AllOptions, Option, WithOptions, WithDefaultOptions
from labrea.template import Template
from labrea.types import Value
import pytest


def test_basic():
    option = Option('A')
    options = {'A': 42}

    assert option.evaluate(options) == 42
    option.validate(options)
    assert option.keys(options) == {'A'}
    assert option.explain(options) == {'A'}
    assert option.explain() == {'A'}


def test_multiple_provided():
    option = Option('A')
    options = {'A': 42, 'V': 43}

    assert option.evaluate(options) == 42
    option.validate(options)
    assert option.keys(options) == {'A'}
    assert option.explain(options) == {'A'}


def test_missing():
    option = Option('A')
    options = {'V': 42}

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.evaluate(options)
        assert excinfo.value.key == 'A'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.validate(options)
        assert excinfo.value.key == 'A'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.keys(options)
        assert excinfo.value.key == 'A'

    assert option.explain(options) == {'A'}


def test_default():
    option = Option('A', default=42)
    missing = {}
    provided = {'A': 43}

    assert option.evaluate(missing) == 42
    option.validate(missing)
    assert option.keys(missing) == set()
    assert option.explain(missing) == set()

    assert option.evaluate(provided) == 43
    option.validate(provided)
    assert option.keys(provided) == {'A'}
    assert option.explain(provided) == {'A'}


def test_default_factory():
    option = Option('A', default_factory=lambda: 42)
    missing = {}
    provided = {'A': 43}

    assert option.evaluate(missing) == 42
    option.validate(missing)
    assert option.keys(missing) == set()
    assert option.explain(missing) == set()

    assert option.evaluate(provided) == 43
    option.validate(provided)
    assert option.keys(provided) == {'A'}
    assert option.explain(provided) == {'A'}


def test_evaluatable_default():
    option = Option('A', default=Option('V'))
    options = {'V': 42}

    assert option.evaluate(options) == 42
    option.validate(options)
    assert option.keys(options) == {'V'}
    assert option.explain(options) == {'V'}
    assert option.explain() == {'V'}


def test_nested():
    option = Option('A.V')
    options = {'A': {'V': 42}}

    assert option.evaluate(options) == 42
    option.validate(options)
    assert option.keys(options) == {'A.V'}
    assert option.explain(options) == {'A.V'}


def test_nested_missing():
    option = Option('A.V')
    options = {'A': {}}

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.evaluate(options)
        assert excinfo.value.key == 'A.V'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.validate(options)
        assert excinfo.value.key == 'A.V'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.keys(options)
        assert excinfo.value.key == 'A.V'

    assert option.explain(options) == {'A.V'}


def test_template_default():
    option = Option('A', default='{V}.{C}')
    assert isinstance(option.default, Template)


def test_confectioner_templating():
    option = Option('A')
    present = {'A': '{V} {C}!', 'V': 'Hello', 'C': 'World'}
    missing = {'A': '{V} {C}!', 'V': 'Hello'}

    assert option.evaluate(present) == 'Hello World!'
    option.validate(present)
    assert option.keys(present) == {'A', 'V', 'C'}

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.evaluate(missing)
        assert excinfo.value.key == 'C'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.validate(missing)
        assert excinfo.value.key == 'C'

    with pytest.raises(KeyNotFoundError) as excinfo:
        option.keys(missing)
        assert excinfo.value.key == 'C'

    assert option.explain(present) == {'A', 'V', 'C'}
    assert option.explain(missing) == {'A', 'V', 'C'}
    assert option.explain() == {'A'}


def test_repr():
    assert repr(Option('A')) == "Option('A')"
    assert repr(Option('A', default=42)) == "Option('A', default=Value(42))"
    assert repr(Option('A', default='{V}.{C}')) == f"Option('A', default={repr(Template('{V}.{C}'))})"
    assert repr(Option('A', doc='Hello, World!')) == "Option('A')"


def test_with_options():
    w = WithOptions(Option('A'), {'A': 42})

    assert w.evaluate({}) == 42
    assert w.evaluate({'A': 43}) == 42

    w.validate({})
    w.validate({'A': 43})

    assert w.keys({}) == set()
    assert w.keys({'A': 43}) == set()

    assert w.explain({}) == set()
    assert w.explain({'A': 43}) == set()

    assert repr(w) == "WithOptions(Option('A'), {'A': 42})"


def test_with_default_options():
    w = WithDefaultOptions(Option('A'), {'A': 42})

    assert w.evaluate({}) == 42
    assert w.evaluate({'A': 43}) == 43

    w.validate({})
    w.validate({'A': 43})

    assert w.keys({}) == set()
    assert w.keys({'A': 43}) == {'A'}

    assert w.explain({}) == set()
    assert w.explain({'A': 43}) == {'A'}

    assert repr(w) == "WithDefaultOptions(Option('A'), {'A': 42})"


def test_all_options():
    options = {'A': 'X', 'B': 'Y', 'C': '{A}/{B}'}
    resolved = {'A': 'X', 'B': 'Y', 'C': 'X/Y'}

    assert AllOptions.evaluate(options) == resolved
    AllOptions.validate(options)
    assert AllOptions.keys(options) == {'A', 'B', 'C'}
    assert AllOptions.explain(options) == {'A', 'B', 'C'}
    assert AllOptions.explain() == set()
    assert repr(AllOptions) == "AllOptions"


def test_all_options_cannot_resolve():
    options = {"A": "{B}"}
    with pytest.raises(KeyNotFoundError) as excinfo:
        AllOptions.evaluate(options)
        assert excinfo.value.key == 'B'
    with pytest.raises(KeyNotFoundError) as excinfo:
        AllOptions.validate(options)
        assert excinfo.value.key == 'B'


def test_namespace_full():
    @Option.namespace("PKG")
    class PKG:
        A: str

    inner = {"A": "a"}
    options = {"PKG": inner}

    assert PKG(options) == inner
    PKG.validate(options)
    assert PKG.keys(options) == {"PKG"}
    assert PKG.explain(options) == {"PKG"}
    assert PKG.explain() == set()

    assert repr(PKG) == "Namespace('PKG')"


def test_namespace_inferred():
    @Option.namespace
    class PKG:
        A: str
        class MODULE:
            B: int

    assert PKG.A({"PKG": {"A": "a"}}) == "a"
    assert PKG.MODULE.B({"PKG": {"MODULE": {"B": 1}}}) == 1


def test_namespace_explicit():
    @Option.namespace("MY-PKG")
    class PKG:
        A = Option("X")
        @Option.namespace("MY-MODULE")
        class MODULE:
            B = Option("Y")

    assert PKG.A({"MY-PKG": {"X": "a"}}) == "a"
    assert PKG.MODULE.B({"MY-PKG": {"MY-MODULE": {"Y": 1}}}) == 1


def test_namespace_auto():
    @Option.namespace
    class PKG:
        A = Option.auto(doc="A as string") >> str
        B = Option.auto(doc="B")

    assert PKG.A({"PKG": {"A": 1}}) == "1"
    assert PKG.A.__doc__ == "A as string"
    assert PKG.B.__doc__ == "B"


def test_namespace_default():
    @Option.namespace
    class PKG:
        A: str = "a"
        B = Value(1) >> str

    assert PKG.A() == "a"
    assert PKG.B() == "1"

    with pytest.raises(TypeError):
        @Option.namespace
        class _:
            A = object()



def test_namespace_doc():
    @Option.namespace
    class PKG:
        A: str
        class MODULE:
            B = Option.auto(1, doc="B")

    assert PKG.__doc__ == 'Namespace PKG:\n  Option PKG.A\n  Namespace PKG.MODULE:\n    Option PKG.MODULE.B (default 1): B'



from typing import Callable, Optional

from mypy.plugin import ClassDefContext, Plugin
from mypy.plugins.common import Instance, add_attribute_to_class  # type: ignore

DATASETCLASS_DECORATOR_FULLNAME = "labrea.datasetclass.datasetclass"
INTERFACE_DECORATOR_FULLNAME = "labrea.interface.interface"
IMPLEMENTATION_DECO_TYPE_FULLNAME = "labrea.interface._ImplDecoProto"
EVALUATABLE_TYPE_FULLNAME = "labrea.evaluatable.Evaluatable"


class LabreaPlugin(Plugin):
    def get_class_decorator_hook_2(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], bool]]:
        if fullname == INTERFACE_DECORATOR_FULLNAME:
            return self.interface_callback
        if fullname == DATASETCLASS_DECORATOR_FULLNAME:
            return self.datasetclass_callback

        return None

    @staticmethod
    def interface_callback(ctx: ClassDefContext):
        for name, attribute in ctx.cls.info.names.copy().items():
            if name.startswith("_"):
                continue
            add_attribute_to_class(  # type: ignore [call-arg]
                api=ctx.api,
                cls=ctx.cls,
                name=name,
                typ=ctx.api.named_type_or_none(  # type: ignore [arg-type]
                    EVALUATABLE_TYPE_FULLNAME
                ),
                overwrite_existing=True,
            )

        add_attribute_to_class(
            api=ctx.api,
            cls=ctx.cls,
            name="implementation",
            typ=ctx.api.named_type_or_none(  # type: ignore [arg-type]
                IMPLEMENTATION_DECO_TYPE_FULLNAME
            ),
        )

        return True

    @staticmethod
    def datasetclass_callback(ctx: ClassDefContext):
        add_attribute_to_class(
            api=ctx.api, cls=ctx.cls, name="result", typ=Instance(ctx.cls.info, [])
        )
        return True


def plugin(version: str):
    # ignore version argument if the plugin works with all mypy versions.
    return LabreaPlugin



import pytest
import threading
from labrea.runtime import Request, handle, inherit


class UnhandledRequest(Request):
    def __init__(self, value):
        self.value = value


class EchoRequest(Request[str]):
    message: str

    def __init__(self, message: str):
        self.message = message


@EchoRequest.handle
def echo(request: EchoRequest) -> str:
    return request.message


def void(request) -> None:
    pass


def test_request():
    assert EchoRequest("hello").run() == "hello"

    with pytest.raises(TypeError):
        UnhandledRequest("world").run()


def test_handle():
    assert EchoRequest("hello").run() == "hello"

    with handle(EchoRequest, void):
        assert EchoRequest("hello").run() is None

    with handle(EchoRequest, void):
        with handle(EchoRequest, echo):
            assert EchoRequest("hello").run() is "hello"

    with handle({EchoRequest: void}):
        assert EchoRequest("hello").run() is None

    with handle({EchoRequest: void, UnhandledRequest: void}):
        assert EchoRequest("hello").run() is None
        assert UnhandledRequest("world").run() is None

    assert EchoRequest("hello").run() == "hello"

    with pytest.raises(TypeError):
        with handle(42, void):
            pass

    with pytest.raises(TypeError):
        with handle(EchoRequest, 42):
            pass


def test_inherit():
    def worker(parent: threading.Thread, request: Request):
        inherit(parent)
        return request.run()

    with pytest.raises(TypeError):
        threading.Thread(target=worker, args=(threading.current_thread(), UnhandledRequest("hello"))).run()

    with handle(UnhandledRequest, void):
        threading.Thread(target=worker, args=(threading.current_thread(), UnhandledRequest("hello"))).run()



import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from typing import Dict, Generic, Optional, Set, Tuple

from .types import Evaluatable, MaybeEvaluatable, Options

P = ParamSpec("P")


class Arguments(Generic[P]):
    """A class representing the arguments of a function or method.

    Arguments
    ---------
    *args : P.args
        The positional arguments.
    **kwargs : P.kwargs
        The keyword arguments.
    """

    args: P.args
    kwargs: P.kwargs

    def __init__(self, *args: P.args, **kwargs: P.kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self) -> str:
        args_repr = ", ".join(map(repr, self.args))
        kwargs_repr = ", ".join(
            f"{key}={value!r}" for key, value in self.kwargs.items()
        )
        return f"Arguments({', '.join(filter(bool, [args_repr, kwargs_repr]))})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Arguments):
            return NotImplemented
        return self.args == other.args and self.kwargs == other.kwargs


class EvaluatableArgs(Generic[P], Evaluatable["P.args"]):
    """A class representing a set of arguments that can be evaluated.

    Arguments
    ---------
    *args : Evaluatable[P.args]
        The arguments to evaluate.
    """

    args: Tuple[Evaluatable, ...]

    def __init__(self, *args: Evaluatable["P.args"]):
        self.args = args

    def evaluate(self, options: Options) -> "P.args":
        return tuple(arg.evaluate(options) for arg in self.args)  # type: ignore

    def validate(self, options: Options) -> None:
        for arg in self.args:
            arg.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return set().union(*(arg.keys(options) for arg in self.args))

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return set().union(*(arg.explain(options) for arg in self.args))

    def __repr__(self) -> str:
        return f"EvaluatableArgs({', '.join(map(repr, self.args))})"


class EvaluatableKwargs(Generic[P], Evaluatable["P.kwargs"]):
    """A class representing a set of keyword arguments that can be evaluated.

    Arguments
    ---------
    **kwargs : Evaluatable[P.kwargs]
        The keyword arguments to evaluate.
    """

    kwargs: Dict[str, Evaluatable]

    def __init__(self, **kwargs: Evaluatable["P.kwargs"]):
        self.kwargs = kwargs

    def evaluate(self, options: Options) -> "P.kwargs":
        return {key: value.evaluate(options) for key, value in self.kwargs.items()}  # type: ignore

    def validate(self, options: Options) -> None:
        for value in self.kwargs.values():
            value.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return set().union(*(value.keys(options) for value in self.kwargs.values()))

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return set().union(*(value.explain(options) for value in self.kwargs.values()))

    def __repr__(self) -> str:
        return (
            f"EvaluatableKwargs("
            f"{', '.join(f'{key}={value!r}' for key, value in self.kwargs.items())}"
            f")"
        )


class EvaluatableArguments(Evaluatable[Arguments[P]]):
    """A class representing a set of arguments that can be evaluated.

    Arguments
    ---------
    *args : MaybeEvaluatable[P.args]
        The positional arguments to evaluate.
    **kwargs : MaybeEvaluatable[P.kwargs]
        The keyword arguments to evaluate
    """

    args: EvaluatableArgs[P]
    kwargs: EvaluatableKwargs[P]

    def __init__(self, *args: Evaluatable["P.args"], **kwargs: Evaluatable["P.kwargs"]):
        self.args = EvaluatableArgs(*args)
        self.kwargs = EvaluatableKwargs(**kwargs)

    def evaluate(self, options: Options) -> Arguments[P]:
        return Arguments(*self.args.evaluate(options), **self.kwargs.evaluate(options))

    def validate(self, options: Options) -> None:
        self.args.validate(options)
        self.kwargs.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return self.args.keys(options).union(self.kwargs.keys(options))

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return self.args.explain(options) | self.kwargs.explain(options)

    def __repr__(self) -> str:
        return (
            f"EvaluatableParameters("
            f"{', '.join(map(repr, self.args.args))}, "
            f"{', '.join(f'{key}={value!r}' for key, value in self.kwargs.kwargs.items())}"
            f")"
        )


def arguments(
    *args: MaybeEvaluatable["P.args"], **kwargs: MaybeEvaluatable["P.kwargs"]
) -> Evaluatable[Arguments[P]]:
    """Create an EvaluatableArguments object from the given arguments.

    Arguments
    ---------
    *args : MaybeEvaluatable[P.args]
        The positional arguments to evaluate.
    **kwargs : MaybeEvaluatable[P.kwargs]
        The keyword arguments to evaluate
    """
    return EvaluatableArguments(
        *(Evaluatable.ensure(arg) for arg in args),
        **{key: Evaluatable.ensure(value) for key, value in kwargs.items()},
    )



import logging
from typing import Optional, Set, TypeVar

from . import runtime
from .computation import Effect
from .option import Option
from .runtime import Request
from .types import Evaluatable, Options

A = TypeVar("A")


class LogRequest(Request[None]):
    """A request to log a message.

    Arguments
    ---------
    level : int
        The logging level (see the :mod:`logging` module).
    name : str
        The name of the logger; usually the module name.
    msg : str
        The message to log.
    options : Options
        The options used during evaluation
    """

    level: int
    name: str
    msg: str
    options: Options

    def __init__(self, level: int, name: str, msg: str, options: Options):
        self.options = options
        self.level = level
        self.name = name
        self.msg = msg


def CRITICAL(name: str, msg: str, options: Options) -> None:
    """Log a message at the CRITICAL level."""
    return LogRequest(logging.CRITICAL, name, msg, options).run()


def ERROR(name: str, msg: str, options: Options) -> None:
    """Log a message at the ERROR level."""
    return LogRequest(logging.ERROR, name, msg, options).run()


def WARNING(name: str, msg: str, options: Options) -> None:
    """Log a message at the WARNING level."""
    return LogRequest(logging.WARNING, name, msg, options).run()


def INFO(name: str, msg: str, options: Options) -> None:
    """Log a message at the INFO level."""
    return LogRequest(logging.INFO, name, msg, options).run()


def DEBUG(name: str, msg: str, options: Options) -> None:
    """Log a message at the DEBUG level."""
    return LogRequest(logging.DEBUG, name, msg, options).run()


def _disabled_logging_handler(request: LogRequest) -> None:
    pass


@LogRequest.handle
def _builtin_logging_handler(request: LogRequest) -> None:
    if Option("LABREA.LOGGING.DISABLED", False)(request.options):
        return _disabled_logging_handler(request)

    logging.getLogger(request.name).log(request.level, request.msg)


def disabled() -> runtime.Runtime:
    """Return a runtime that disables logging.

    Can be used as a context manager to disable logging for the duration of the context block.


    Example Usage
    -------------
    >>> import labrea.logging
    >>> with labrea.logging.disabled():
    ...     pass
    """
    return runtime.handle(LogRequest, _disabled_logging_handler)


class LogEffect(Effect):
    """An effect that logs a message.

    Arguments
    ---------
    level : int
        The logging level (see the :mod:`logging` module).
    name : str
        The name of the logger; usually the module name.
    msg : str
        The message to log.
    """

    level: int
    name: str
    msg: str

    def __init__(self, level: int, name: str, msg: str):
        self.level = level
        self.name = name
        self.msg = msg

    def transform(self, value: None, options: Optional[Options] = None) -> None:
        """Log the message."""
        return LogRequest(self.level, self.name, self.msg, options or {}).run()

    def validate(self, options: Options) -> None:
        """Always validates"""
        pass

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return an empty set, as no options are required."""
        return set()


class Logged(Evaluatable[A]):
    """An Evaluatable that logs a message before (or after) evaluating another Evaluatable.

    Arguments
    ---------
    evaluatable : Evaluatable[A]
        The Evaluatable to evaluate.
    level : int
        The logging level (see the :mod:`logging` module).
    name : str
        The name of the logger; usually the module name.
    msg : str
        The message to log.
    log_first : bool, optional
        Whether to log the message before or after evaluating the Evaluatable. Default is True.
    """

    evaluatable: Evaluatable[A]
    level: int
    name: str
    msg: str
    log_first: bool

    def __init__(
        self,
        evaluatable: Evaluatable[A],
        level: int,
        name: str,
        msg: str,
        log_first: bool = True,
    ):
        self.evaluatable = evaluatable
        self.level = level
        self.name = name
        self.msg = msg
        self.log_first = log_first

    def _request(self, options: Options) -> LogRequest:
        return LogRequest(self.level, self.name, self.msg, options)

    def evaluate(self, options: Options) -> A:
        """Evaluate the Evaluatable and log the message."""
        if self.log_first:
            self._request(options).run()
            return self.evaluatable.evaluate(options)
        else:
            value = self.evaluatable.evaluate(options)
            self._request(options).run()
            return value

    def validate(self, options: Options) -> None:
        """Validate the Evaluatable."""
        self.evaluatable.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required to evaluate the Evaluatable."""
        return self.evaluatable.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required to evaluate the Evaluatable."""
        return self.evaluatable.explain(options)

    def __repr__(self) -> str:
        return (
            f"Logged({self.evaluatable!r}, {self.level!r}, "
            f"{self.name!r}, {self.msg!r})"
            if self.log_first
            else f"Logged({self.evaluatable!r}, {self.level!r}, "
            f"{self.name!r}, {self.msg!r}, log_first=False)"
        )



import itertools
from typing import Dict, Iterable, Optional, Set, Tuple, TypeVar

from confectioner.templating import set_dotted_key

from .exceptions import EvaluationError
from .option import WithOptions
from .types import JSON, Evaluatable, MaybeEvaluatable, Options

A = TypeVar("A", covariant=True)


class Iter(Evaluatable[Iterable[A]]):
    """A class representing multiple evaluatables as an iterable.

    This is used to lazily evaluate multiple evaluatables in a single
    operation. The result is an iterable of the evaluated values.

    Arguments
    ----------
    *evaluatables : MaybeEvaluatable[A]
        The evaluatables to evaluate.


    Example Usage
    -------------
    >>> from labrea import Iter, Option
    >>> i = Iter(Option('A.X'), Option('B.Y'))
    >>> for value in i({'A': {'X': 1}, 'B': {'Y': 2}}):
    ...     print(value)
    1
    2
    """

    evaluatables: Iterable[Evaluatable[A]]

    def __init__(self, *evaluatables: Evaluatable[A]):
        self.evaluatables = tuple(Evaluatable.ensure(e) for e in evaluatables)

    def evaluate(self, options: Options) -> Iterable[A]:
        """Evaluate the evaluatables and return an iterable of the results."""
        return (evaluatable.evaluate(options) for evaluatable in self.evaluatables)

    def validate(self, options: Options) -> None:
        """Validate that all evaluatables can be evaluated."""
        for evaluatable in self.evaluatables:
            evaluatable.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the evaluatables."""
        return {
            key
            for evaluatable in self.evaluatables
            for key in evaluatable.keys(options)
        }

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the evaluatables."""
        return {
            explanation
            for evaluatable in self.evaluatables
            for explanation in evaluatable.explain(options)
        }

    def __repr__(self) -> str:
        return f"Iter({', '.join(map(repr, self.evaluatables))})"


class Map(Evaluatable[Iterable[Tuple[Dict[str, JSON], A]]]):
    """A class that represents the same evaluatable repeated over multiple options.

    This is used to evaluate a single evaluatable multiple times with different
    options. The result is an iterable of tuples, where the first element is the
    options used to evaluate the evaluatable, and the second element is the
    evaluated value. This is the Labrea equivalent of a for loop.

    Arguments
    ---------
    evaluatable : Evaluatable[A]
        The evaluatable to evaluate.
    iterables : Dict[str, MaybeEvaluatable[Iterable[JSON]]]
        A dictionary mapping option keys to iterables of options. The keys
        can be dotted to represent nested options.


    Example Usage
    -------------
    >>> from labrea import Map, Option, dataset
    >>> @dataset
    ... def hypotenuse(a: int = Option('A'), b: int = Option('B')) -> float:
    ...     return (a ** 2 + b ** 2) ** 0.5
    >>>
    >>> hypotenuses = Map(hypotenuse, {'A': Option('X'), 'B': Option('Y')})
    >>> for options, value in hypotenuses({'X': [3, 4], 'Y': [4, 5]}):
    ...     print(options, value)
    {'A': 3, 'B': 4} 5.0
    {'A': 3, 'B': 5} 5.830951894845301
    {'A': 4, 'B': 4} 5.656854249492381
    {'A': 4, 'B': 5} 6.403124237432848
    """

    evaluatable: Evaluatable[A]
    iterables: Dict[str, Evaluatable[Iterable[JSON]]]

    def __init__(
        self,
        evaluatable: Evaluatable[A],
        iterables: Dict[str, MaybeEvaluatable[Iterable[JSON]]],
    ) -> None:
        self.evaluatable = evaluatable
        self.iterables = {
            key: Evaluatable.ensure(value) for key, value in iterables.items()
        }

    def evaluate(self, options: Options) -> Iterable[Tuple[Dict[str, JSON], A]]:
        """Evaluate the evaluatable with each set of options."""
        return self._iter(options).evaluate(options)

    def validate(self, options: Options) -> None:
        """Validate that both the option iterables and the evaluatable can be evaluated."""
        self._iter(options).validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the option keys required by the evaluatable and the option iterables"""
        return set().union(
            self._iter(options).keys(options),
            *(iterable.keys(options) for iterable in self.iterables.values()),
        )

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the option keys required by the evaluatable and the option iterables"""
        try:
            return set().union(
                self._iter(options or {}).explain(options),
                *(iterable.explain(options) for iterable in self.iterables.values()),
            )
        except EvaluationError:
            return (
                self.evaluatable.explain(options) - self.iterables.keys()
            ) | set().union(
                *(iterable.explain(options) for iterable in self.iterables.values())
            )

    def _iter(
        self, options: Options
    ) -> Evaluatable[Iterable[Tuple[Dict[str, JSON], A]]]:
        return Iter(
            *(
                Iter(
                    Iter(*option_tuples).apply(dict),  # type: ignore
                    WithOptions(  # type: ignore
                        self.evaluatable, self._create_option_set(*option_tuples)
                    ),
                ).apply(lambda x: tuple(x))
                for option_tuples in self._iterate_over_options(options)
            )
        )

    @staticmethod
    def _create_option_set(*args: Tuple[str, JSON]) -> Options:
        options: Dict[str, JSON] = {}
        for key, value in args:
            set_dotted_key(key, value, options)

        return options

    def _iterate_over_options(
        self, options: Options
    ) -> Tuple[Tuple[Tuple[str, JSON], ...], ...]:
        return tuple(
            tuple(zip(self.iterables.keys(), values))
            for values in itertools.product(
                *(iterable.evaluate(options) for iterable in self.iterables.values())
            )
        )

    def __repr__(self) -> str:
        return f"Iterate({self.evaluatable!r}, {self.iterables!r})"

    @property
    def values(self) -> Evaluatable[Iterable[A]]:
        """Strip the options from the results.

        This is a convenience method that returns an evaluatable that
        returns the second element of the tuples returned by the Map
        evaluatable.
        """
        return self.apply(lambda items: (item[1] for item in items))



from . import _version, exceptions, types
from .cache import cached
from .coalesce import Coalesce, coalesce
from .collections import (
    DatasetDict,
    DatasetList,
    DatasetSet,
    DatasetTuple,
    evaluatable_dict,
    evaluatable_list,
    evaluatable_set,
    evaluatable_tuple,
)
from .conditional import Switch, case, switch
from .dataset import abstractdataset, dataset
from .datasetclass import datasetclass
from .interface import implements, interface
from .iterable import Iter, Map
from .option import AllOptions, Option, WithDefaultOptions, WithOptions
from .overload import Overloaded
from .pipeline import pipeline_step
from .template import Template
from .types import Value

__version__ = _version.__version__


__all__ = [
    "Coalesce",
    "DatasetDict",
    "DatasetList",
    "DatasetSet",
    "DatasetTuple",
    "Iter",
    "Map",
    "AllOptions",
    "Option",
    "WithDefaultOptions",
    "WithOptions",
    "Overloaded",
    "Switch",
    "Template",
    "Value",
    "abstractdataset",
    "cached",
    "case",
    "coalesce",
    "dataset",
    "datasetclass",
    "evaluatable_dict",
    "evaluatable_list",
    "evaluatable_set",
    "evaluatable_tuple",
    "exceptions",
    "implements",
    "interface",
    "pipeline_step",
    "switch",
    "types",
]



from labrea.option import Option
import labrea.functions as lf


def test_partial():
    p = Option('A') >> lf.partial(lambda x, y: x + y, 1)

    assert p({'A': 2}) == 3


def test_map():
    m = Option('A') >> lf.map(str) >> list

    assert m({'A': [1, 2]}) == ['1', '2']


def test_filter():
    f = Option('A') >> lf.filter(lambda x: x % 2 == 0) >> list

    assert f({'A': [1, 2, 3, 4]}) == [2, 4]


def test_reduce():
    r1 = Option('A') >> lf.reduce(lambda x, y: x + y)
    r2 = Option('A') >> lf.reduce(lambda x, y: x + y, 10)

    assert r1({'A': [1, 2, 3, 4]}) == 10
    assert r2({'A': [1, 2, 3, 4]}) == 20


def test_into():
    i = Option('A') >> lf.into(lf.partial(lambda x, y, z: x + y + z, z=3))

    assert i({'A': [1, 2]}) == 6
    assert i({'A': {'x': 1, 'y': 2}}) == 6


def test_flatmap():
    f = Option('A') >> lf.flatmap(lambda x: [x, x]) >> list

    assert f({'A': [1, 2]}) == [1, 1, 2, 2]


def test_map_items():
    f = Option('A') >> dict >> lf.map_items(lambda k, v: (k + 1, v + 1)) >> dict

    assert f({'A': [[1, 1], [2, 2]]}) == {2: 2, 3: 3}


def test_map_keys():
    f = Option('A') >> dict >> lf.map_keys(lambda k: k + 1) >> dict

    assert f({'A': [[1, 1], [2, 2]]}) == {2: 1, 3: 2}


def test_map_values():
    f = Option('A') >> dict >> lf.map_values(lambda v: v + 1) >> dict

    assert f({'A': [[1, 1], [2, 2]]}) == {1: 2, 2: 3}


def test_filter_items():
    f = Option('A') >> dict >> lf.filter_items(lambda k, v: (k * v) % 2 == 0) >> dict

    assert f({'A': [[1, 1], [2, 2], [3, 3], [4, 4]]}) == {2: 2, 4: 4}


def test_filter_keys():
    f = Option('A') >> dict >> lf.filter_keys(lambda k: k % 2 == 0) >> dict

    assert f({'A': [[1, 1], [2, 2], [3, 3], [4, 4]]}) == {2: 2, 4: 4}


def test_filter_values():
    f = Option('A') >> dict >> lf.filter_values(lambda v: v % 2 == 0) >> dict

    assert f({'A': [[1, 1], [2, 2], [3, 3], [4, 4]]}) == {2: 2, 4: 4}



import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import functools
import inspect
from typing import Callable, Dict, Generic, Optional, Set, TypeVar, Union, overload

from .arguments import Arguments, arguments
from .types import Evaluatable, MaybeEvaluatable, Options

P = ParamSpec("P")
A = TypeVar("A", covariant=True)
X = TypeVar("X")


class FunctionApplication(Generic[P, A], Evaluatable[A]):
    """A class representing the application of a function to a set of arguments.

    This class is what is used by the :class:`labrea.dataset.Dataset` under the hood.

    Arguments
    ---------
    __func : Callable[P, A]
        The function to apply.
    *args : MaybeEvaluatable[P.args]
        The positional arguments to evaluate.
    **kwargs : MaybeEvaluatable[P.kwargs]
        The keyword arguments to evaluate.
    """

    func: Evaluatable[Callable[P, A]]
    arguments: Evaluatable[Arguments[P]]
    _repr: str

    def __init__(
        self,
        __func: MaybeEvaluatable[Callable[P, A]],
        /,
        *args: "MaybeEvaluatable[P.args]",
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ):
        self.func = Evaluatable.ensure(__func)
        self.arguments = arguments(*args, **kwargs)

        self._repr = (
            f"FunctionApplication("
            f"{getattr(__func, '__name__', repr(__func))}, "
            f"{', '.join(map(repr, args))}"
            f"{', '.join(f'{key}={value!r}' for key, value in kwargs.items())}"
            f")"
        )

    def evaluate(self, options: Options) -> A:
        func = self.func.evaluate(options)
        args = self.arguments.evaluate(options)
        return func(*args.args, **args.kwargs)

    def validate(self, options: Options) -> None:
        self.func.validate(options)
        self.arguments.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return self.func.keys(options) | self.arguments.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return self.func.explain(options) | self.arguments.explain(options)

    def __repr__(self) -> str:
        return self._repr

    @overload
    @classmethod
    def lift(
        cls,
        __func: Callable[P, A],
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> "FunctionApplication[P, A]": ...  # pragma: no cover

    @overload
    @classmethod
    def lift(
        cls,
        __func: None = None,
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> Callable[
        [Callable[P, A]], "FunctionApplication[P, A]"
    ]: ...  # pragma: no cover

    @classmethod
    def lift(
        cls,
        __func: Optional[Callable[P, A]] = None,
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> Union[
        "FunctionApplication[P, A]",
        Callable[[Callable[P, A]], "FunctionApplication[P, A]"],
    ]:
        """Lift a function definition with Evaluatable default arguments to a FunctionApplication.

        This is used by the :func:`labrea.dataset` decorator under the hood. Can be used
        as a decorator with or without keyword arguments for the defaults.

        Arguments
        ---------
        __func : Callable[P, A]
            The function to apply.
        **kwargs : MaybeEvaluatable[P.kwargs]
            Default values for the keyword arguments of the function.

        Returns
        -------
        FunctionApplication[P, A]
            The lifted function application.


        Example Usage
        -------------
        >>> @FunctionApplication.lift
        ... def a_squared(a: int = Option('A')) -> int:
        ...     return a ** 2
        >>>
        >>> @FunctionApplication(a=Option('A'))
        ... def a_squared(a: int) -> int:
        ...     return a ** 2
        >>>
        >>> a_squared = FunctionApplication.lift(lambda a: a**2, a=Option('A'))
        """
        if __func is None:
            return lambda f: cls.lift(f, **kwargs)

        signature = inspect.signature(__func)
        eval_kwargs: Dict[str, Evaluatable["P.kwargs"]] = {}

        for param in signature.parameters.values():
            default = kwargs.get(param.name, param.default)
            if default is param.empty:
                raise TypeError(
                    f"Cannot lift function {__func} with non-defaulted parameters"
                )
            eval_kwargs[param.name] = Evaluatable.ensure(default)

        return FunctionApplication(__func, **eval_kwargs)


class PartialApplication(Generic[P, A], Evaluatable[Callable[..., A]]):
    """A class representing the partial application of a function to a set of arguments.

    This class is used by the :func:`labrea.pipeline_step` decorator under the hood.

    Arguments
    ---------
    __func : Callable[P, A]
        The function to apply.
    *args : MaybeEvaluatable[P.args]
        The positional arguments to evaluate.
    **kwargs : MaybeEvaluatable[P.kwargs]
        The keyword arguments to evaluate.
    """

    func: Evaluatable[Callable[P, A]]
    arguments: Evaluatable[Arguments[P]]
    _repr: str

    def __init__(
        self,
        __func: MaybeEvaluatable[Callable[P, A]],
        /,
        *args: MaybeEvaluatable["P.args"],
        **kwargs: MaybeEvaluatable["P.kwargs"],
    ):
        self.func = Evaluatable.ensure(__func)
        self.arguments = arguments(*args, **kwargs)

        self._repr = (
            f"PartialApplication("
            f"{getattr(__func, '__name__', repr(__func))}, "
            f"{', '.join(map(repr, args))}"
            f"{', '.join(f'{key}={value!r}' for key, value in kwargs.items())}"
            f")"
        )

    def evaluate(self, options: Options) -> Callable[..., A]:
        func = self.func.evaluate(options)
        args = self.arguments.evaluate(options)
        return functools.partial(func, *args.args, **args.kwargs)

    def validate(self, options: Options) -> None:
        self.func.validate(options)
        self.arguments.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return self.func.keys(options) | self.arguments.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return self.func.explain(options) | self.arguments.explain(options)

    def __repr__(self) -> str:
        return self._repr

    @overload
    @classmethod
    def lift(
        cls,
        __func: Callable[P, A],
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> "PartialApplication[P, A]": ...  # pragma: no cover

    @overload
    @classmethod
    def lift(
        cls,
        __func: None = None,
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> Callable[[Callable[P, A]], "PartialApplication[P, A]"]: ...  # pragma: no cover

    @classmethod
    def lift(
        cls,
        __func: Optional[Callable[P, A]] = None,
        /,
        **kwargs: "MaybeEvaluatable[P.kwargs]",
    ) -> Union[
        "PartialApplication[P, A]",
        Callable[[Callable[P, A]], "PartialApplication[P, A]"],
    ]:
        """Lift a function definition with Evaluatable default arguments to a PartialApplication.

        This is used by the :func:`labrea.pipeline_step` decorator under the hood. Can be used
        as a decorator with or without keyword arguments for the defaults.

        Arguments
        ---------
        __func : Callable[P, A]
            The function to apply.
        **kwargs : MaybeEvaluatable[P.kwargs]
            Default values for the keyword arguments of the function.

        Returns
        -------
        PartialApplication[P, A]
            The lifted partial application.


        Example Usage
        -------------
        >>> @PartialApplication.lift
        ... def a_plus_b(a: int, b: int = Option('B')) -> int:
        ...     return a + b
        >>>
        >>> @PartialApplication(b=Option('B'))
        ... def a_plus_b(a: int, b: int) -> int:
        ...     return a + b
        >>>
        >>> a_plus_b = PartialApplication.lift(lambda a, b: a + b, b=Option('B'))
        """
        if __func is None:
            return lambda f: cls.lift(f, **kwargs)

        signature = inspect.signature(__func)
        eval_kwargs: Dict[str, Evaluatable["P.kwargs"]] = {}

        for i, param in enumerate(signature.parameters.values()):
            default = kwargs.get(param.name, param.default)
            if default is not param.empty:
                eval_kwargs[param.name] = Evaluatable.ensure(default)

        return PartialApplication(__func, **eval_kwargs)



from types import FunctionType
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .application import FunctionApplication
from .dataset import Dataset, abstractdataset, dataset
from .option import Option
from .types import Evaluatable, Value

T = TypeVar("T", bound=Type)


def interface(dispatch: Union[Evaluatable[Hashable], str]) -> Callable[[T], "T"]:
    """A decorator that creates an interface from a class.

    An interface is a class that defines a set of members that must be implemented
    by its implementations. The interface is defined by the members of the class,
    which are abstract datasets. The interface must have a dispatch, which is an
    evaluatable that returns a key that is used to determine which implementation.

    Downstream datasets can use interface members as if they were normal datasets.
    This allows developers to define a set of related datasets that must be implemented
    for a particular use case, enabling polymorphism and dependency injection.

    Arguments
    ----------
    dispatch : Evaluatable[Hashable] or str
        The evaluatable that determines which implementation to use. This can be
        an evaluatable (like an Option) or a string representing an Option.

    Returns
    -------
    Callable[[T], T]
        A decorator that creates an interface from a class.


    Example Usage
    -------------
    >>> from labrea import interface, abstractdataset, dataset, Option
    >>> @interface("DISPATCH.KEY")
    ... class MyInterface:
    ...     a: float  # Implicit member
    ...
    ...     # Explicit member
    ...     @abstractdataset
    ...     def b() -> str:
    ...         pass
    ...
    ...     # Member with default implementation
    ...     @dataset
    ...     def c() -> str:
    ...         return 'C'
    ...
    ...     # Member with default implementation, not using @dataset
    ...     d: str = Option('D')
    >>>
    >>> @dataset
    ... def a_squared(a: float = MyInterface.a) -> float:
    ...     return a ** 2
    >>>
    """
    if isinstance(dispatch, type):
        raise TypeError(
            "@interface requires that an Evaluatable (or str representing an Option) be provided "
            "to dispatch over, but a type was provided. Did you forget to provide a dispatch? "
            '(e.g. @interface instead of @interface("DISPATCH.KEY"))?'
        )

    if isinstance(dispatch, str):
        dispatch = Option(dispatch)

    def wrapper(cls):
        return Interface(cls.__name__, cls.__bases__, dict(cls.__dict__), dispatch)

    return wrapper


def implements(
    *interfaces: "Interface", alias: Union[Hashable, List[Hashable], None] = None
) -> Callable[[T], T]:
    """Decorator used for implementing one or more interfaces at once.

    Arguments
    ---------
    *interfaces : Interface
        The interfaces to implement.
    alias : Union[Hashable, List[Hashable], None]
        The alias or aliases to register the implementation under.
        Keyword-only argument; required.

    Returns
    -------
    Callable[[T], T]
        A decorator that creates an implementation from a class.


    Example Usage
    -------------
    >>> @implements(Interface1, Interface2, alias="ALIAS")
    ... class MyImplementation:
    ...     # Static value member
    ...     a: str = 'A'
    ...     # Non-dataset Evaluatable member
    ...     b: str = Option('B')
    ...     # Explicit dataset member
    ...     @dataset
    ...     def c() -> str:
    ...         return 'C'
    ...     # Implicit dataset member
    ...     def d() -> str:
    ...         return 'D'
    """
    if alias is None:
        raise ValueError(
            "The @implements decorator requires at least one alias to be provided."
        )
    aliases = tuple(alias) if isinstance(alias, list) else (alias,)

    def wrapper(cls):
        return Implementation(
            cls.__name__, cls.__bases__, dict(cls.__dict__), interfaces, aliases
        )

    return wrapper


class Interface(type):
    """Metaclass for interfaces.

    Interfaces are classes that define a set of members that must be implemented by
    its implementations. The interface is defined by the members of the class, which
    are abstract datasets. The interface must have a dispatch, which is an evaluatable
    that returns a key that is used to determine which implementation to use.

    Members of an interface are defined by the class attributes and annotations. Members
    can be defined implicitly using a type annotation, or explicitly using the
    :code:`@abstractdataset` decorator. Members can also have a default implementation.
    Usually this is done by using the :code:`@dataset` decorator, but any class attribute
    that is an Evaluatable can be used as a default implementation.

    Normally, rather than using this metaclass directly, the :code:`@interface` decorator
    should be used to create an interface.
    """

    _dispatch: Evaluatable[Hashable]
    _members: Dict[str, Dataset]

    def __new__(mcs, name, bases, namespace, dispatch: Evaluatable[Hashable]):
        namespace["_dispatch"] = dispatch
        return super().__new__(mcs, name, bases, namespace)

    def __init__(
        cls,
        name: str,
        bases: Tuple[type, ...],
        dct: Dict[str, Any],
        dispatch: Evaluatable[Hashable],
    ):
        super().__init__(name, bases, dct)

        for name, typ in dct.get("__annotations__", {}).items():
            if name.startswith("_") or name in dct:
                continue

            def _abstractdataset():
                pass  # pragma: no cover

            _abstractdataset.__qualname__ = f"{cls.__name__}.{name}"
            _abstractdataset.__name__ = name
            setattr(cls, name, abstractdataset(_abstractdataset, dispatch=dispatch))

        for key, val in dct.items():
            if key.startswith("_"):
                continue

            if isinstance(val, staticmethod):
                val = val.__func__
                setattr(cls, key, val)

            if isinstance(val, FunctionType):
                setattr(cls, key, dataset(val, dispatch=dispatch))
            elif isinstance(val, Dataset):
                val.set_dispatch(dispatch)
            else:

                def _dataset(v=Evaluatable.ensure(val)):
                    return v

                _dataset.__qualname__ = f"{cls.__qualname__}.{key}"
                _dataset.__name__ = key
                setattr(cls, key, dataset(_dataset, dispatch=dispatch))

    def __setattr__(self, key, value):
        if key in ("implementation",):
            raise AttributeError(f"Cannot create a member called {key} on Interface.")

        super().__setattr__(key, value)

    def implementation(cls, alias: Union[Hashable, List[Hashable]]) -> Callable[[T], T]:
        """Decorator for implementing a single interface.

        This is shorthand for using the :code:`@implements` decorator with a single interface.

        Arguments
        ---------
        alias : Union[Hashable, List[Hashable]]
            The alias or aliases to register the implementation under.
        """
        if isinstance(alias, type):
            raise TypeError(
                f"@{cls.__name__}.implementation requires that at least one alias be provided, "
                f"but a type was provided. Did you forget to provide an alias? "
                f"(e.g. @{cls.__name__}.implementation instead of "
                f'@{cls.__name__}.implementation("ALIAS"))?'
            )

        return implements(cls, alias=alias)

    def __repr__(self):
        return f"<Interface {self.__module__}.{self.__qualname__}>"


class Implementation(type):
    """Metaclass for interface implementations.

    Implementations are classes that implement one or more interfaces. Implementations
    are defined by the members of the class, which are datasets or other Evaluatables.
    When the implementation is created, the members are added as overloads to the
    corresponding members of the interfaces. If a member is not implemented, an error
    is raised.

    Normally, rather than using this metaclass directly, the :code:`@implements` decorator
    or the :code:`@MyInterface.implementation` decorator should be used.
    """

    _interfaces: Tuple[Interface, ...]

    def __new__(
        mcs,
        name,
        bases,
        namespace,
        interfaces: Tuple[Interface, ...],
        aliases: Tuple[Hashable, ...],
    ):
        namespace["_interfaces"] = interfaces
        return super().__new__(mcs, name, bases, namespace)

    def __init__(
        cls,
        name,
        bases,
        dct,
        interfaces: Tuple[Interface, ...],
        aliases: Tuple[Hashable, ...],
    ):
        super().__init__(name, bases, dct)

        members = _get_members(interfaces)
        overloads = _build_overloads(dct, members, interfaces)

        for key, member_list in members.items():
            overload = overloads.get(key)

            for member in member_list:
                if member.is_abstract and overload is None:
                    raise TypeError(f"No implementation provided for {member}.")

            if overload is not None:
                for member in member_list:
                    for alias in aliases:
                        member.register(alias, overload)
                setattr(cls, key, overload)

    def __repr__(self):
        return (
            f"<Implementation {self.__module__}.{self.__qualname__} of "
            f'{", ".join(i.__qualname__ for i in self._interfaces)}>'
        )


def _get_members(interfaces: Tuple[Interface, ...]) -> Dict[str, List[Dataset]]:
    members: Dict[str, List[Dataset]] = {}

    for interface_ in interfaces:
        for name, member in interface_.__dict__.items():
            if not name.startswith("_") and isinstance(member, Dataset):
                members.setdefault(name, []).append(member)

    return members


def _build_overloads(
    dct: Dict[str, Any], members: Dict[str, List[Dataset]], interfaces
) -> Dict[str, Evaluatable]:
    overloads: Dict[str, Evaluatable] = {}

    for key, val in dct.items():
        if key.startswith("_"):
            continue

        if key not in members:
            raise TypeError(
                f"No interface in {interfaces} has a member named {key} to overload."
            )

        if isinstance(val, staticmethod):
            val = val.__func__

        overload: Evaluatable
        if isinstance(val, FunctionType):
            overload = FunctionApplication.lift(val)
        elif not isinstance(val, Evaluatable):
            overload = Value(val)
        else:
            overload = val

        overloads[key] = overload

    return overloads


class _ImplDecoProto(Protocol):
    def __call__(self, alias: Hashable, *aliases: Hashable) -> Callable[[T], T]: ...



import pytest

from labrea.computation import Computation, ChainedEffect, CallbackEffect
from labrea.exceptions import EvaluationError
from labrea.option import Option
from labrea.pipeline import pipeline_step


def test_callback_effect():
    store = None

    @pipeline_step
    def store_value_add_a(value: int, a: int = Option('A')) -> None:
        nonlocal store
        store = value + a

    effect = CallbackEffect(store_value_add_a)

    assert effect.transform(1, {'A': 1}) is None
    assert store == 2
    effect.validate({'A': 1})
    assert effect.explain() == {'A'}
    assert repr(effect) == f"CallbackEffect({store_value_add_a!r})"

    with pytest.raises(EvaluationError):
        effect.transform(1, {})
    with pytest.raises(EvaluationError):
        effect.validate({})


def test_chained_effect():
    a = lambda x: None
    b = lambda x: None

    effect = ChainedEffect(CallbackEffect(a), CallbackEffect(b))

    assert effect.transform(1) is None
    effect.validate({})
    assert effect.explain() == set()
    assert repr(effect) == f"ChainedEffect(CallbackEffect(Value({a!r})), CallbackEffect(Value({b!r})))"


def test_computation():
    store = None

    @pipeline_step
    def store_value_add_a(value: int, a: int = Option('A')) -> None:
        nonlocal store
        store = value + a

    comp = Computation(Option('A'), CallbackEffect(store_value_add_a))

    assert comp.evaluate({'A': 1}) == 1
    assert store == 2
    comp.validate({'A': 1})
    assert comp.keys({'A': 1}) == {'A'}
    assert comp.explain() == {'A'}

    assert repr(comp) == f"Computation(Option('A'), CallbackEffect({store_value_add_a!r}))"



import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import builtins
import functools
import itertools
from types import MappingProxyType
from typing import Any, Callable, Hashable, Iterable, Mapping, Tuple, TypeVar, Union

from ._missing import MISSING, MaybeMissing
from .application import PartialApplication
from .pipeline import Pipeline, PipelineStep, pipeline_step
from .types import Evaluatable, MaybeEvaluatable

A = TypeVar("A")
B = TypeVar("B")
K1 = TypeVar("K1", bound=Hashable)
K2 = TypeVar("K2", bound=Hashable)
V1 = TypeVar("V1")
V2 = TypeVar("V2")
P = ParamSpec("P")


def partial(
    __func: MaybeEvaluatable[Callable[P, A]],
    *args: MaybeEvaluatable["P.args"],
    **kwargs: MaybeEvaluatable["P.kwargs"],
) -> Evaluatable[Callable[..., A]]:
    """Analog of functools.partial that works with Evaluatables.

    Arguments
    ----------
    func : MaybeEvaluatable[Callable[P, A]]
        The function to partially apply.
    *args : MaybeEvaluatable[P.args]
        The positional arguments to partially apply to the function. These
        can be Evaluatables that return values, or constant values.
    **kwargs : MaybeEvaluatable[P.kwargs]
        The keyword arguments to partially apply to the function. These can
        be Evaluatables that return values, or constant values.

    Returns
    -------
    Evaluatable[Callable[..., A]]
        An Evaluatable that returns a partially applied function.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.partial(lambda x, y: x + y, y=Option('B')))({'A': 1, 'B': 2})
    3
    """
    return PartialApplication(__func, *args, **kwargs)


def map(
    func: MaybeEvaluatable[Callable[[A], B]],
) -> PipelineStep[Iterable[A], Iterable[B]]:
    """Create a pipeline step that maps a function over an iterable.

    This allows for calling map in a functional style over an iterable
    using the :code:`.apply()` method or the :code:`>>` operator on an
    Evaluatable object that returns an iterable.

    Arguments
    ----------
    func : MaybeEvaluatable[Callable[[A], B]]
        The function to apply to each element of the iterable. This can be
        an Evaluatable that returns a function, or a constant function.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.map(lambda x: x + 1) >> list)({'A': [1, 2, 3, 4]})
    [2, 3, 4, 5]
    """
    return PipelineStep(
        partial(builtins.map, func),  # type: ignore [arg-type]
        f"map({func!r})",
    )


def filter(
    func: MaybeEvaluatable[Callable[[A], bool]],
) -> PipelineStep[Iterable[A], Iterable[A]]:
    """Create a pipeline step that filters an iterable using a predicate function.

    This allows for calling filter in a functional style over an iterable
    using the :code:`.apply()` method or the :code:`>>` operator on an
    Evaluatable object that returns an iterable.

    Arguments
    ----------
    func : MaybeEvaluatable[Callable[[A], bool]]
        The predicate function to apply to each element of the iterable. This
        can be an Evaluatable that returns a function, or a constant function.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.filter(lambda x: x % 2 == 0) >> list)({'A': [1, 2, 3, 4]})
    [2, 4]
    """
    return PipelineStep(
        partial(builtins.filter, func),  # type: ignore [arg-type]
        f"filter({func!r})",
    )


def _reduce(
    func: Callable[[A, B], A], iterable: Iterable[B], initial: MaybeMissing[A]
) -> A:
    if initial is MISSING:
        return functools.reduce(func, iterable)  # type: ignore

    return functools.reduce(func, iterable, initial)


def reduce(
    func: MaybeEvaluatable[Callable[[A, B], A]],
    initial: MaybeMissing[MaybeEvaluatable[A]] = MISSING,
) -> PipelineStep[Iterable[B], A]:
    """Create a pipeline step that reduces an iterable using a binary function.

    This allows for calling reduce in a functional style over an iterable
    using the :code:`.apply()` method or the :code:`>>` operator on an
    Evaluatable object that returns an iterable.

    Arguments
    ----------
    func : MaybeEvaluatable[Callable[[A, A], A]]
        The binary function to apply to each element of the iterable. This
        can be an Evaluatable that returns a function, or a constant function.
    initial : MaybeMissing[MaybeEvaluatable[A]], optional
        The initial value to use for the reduction. If not provided, the
        first element of the iterable is used as the initial value. This
        can be an Evaluatable that returns a value, or a constant value.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.reduce(lambda x, y: x + y))({'A': [1, 2, 3, 4]})
    10
    """
    return PipelineStep(
        partial(_reduce, func, initial=initial),
        f"reduce({func!r})"
        if initial is MISSING
        else f"reduce({func!r}, initial={initial!r})",
    )  # type: ignore [arg-type]


def into(
    func: MaybeEvaluatable[Callable[..., A]],
) -> PipelineStep[Union[Iterable, Mapping[str, Any]], A]:
    """Convert a function that takes positional arguments into one that takes an iterable.

    This can be useful if you have an evaluatable that returns an iterable of args or a mapping
    of kwargs and you want to unpack that into a function that takes positional or keyword
    arguments.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[..., A]]
        The function to apply to the arguments.
        Can be an evaluatable that returns a function or a constant function.

    Returns
    -------
    PipelineStep[Union[Iterable, Mapping[str, Any], A]
        A pipeline step that applies the function to the arguments after unpacking.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.into(lambda x, y: x + y))({'A': [1, 2]})
    3
    >>> (Option('A') >> F.into(lambda x, y: x + y))({'A': {'x': 1, 'y': 2}})
    3
    """

    @pipeline_step
    def _into(
        args: Union[Iterable, Mapping[str, Any]],
        f: Callable[..., A] = func,  # type: ignore[assignment]
    ) -> A:
        return f(**args) if isinstance(args, Mapping) else f(*args)

    return PipelineStep(_into, f"into({func!r})")


def flatmap(
    func: MaybeEvaluatable[Callable[[A], Iterable[B]]],
) -> PipelineStep[Iterable[A], Iterable[B]]:
    """Create a pipeline step that maps a function over an iterable and flattens the result.

    This allows for calling flatmap in a functional style over an iterable
    using the :code:`.apply()` method or the :code:`>>` operator on an
    Evaluatable object that returns an iterable.

    Arguments
    ----------
    func : MaybeEvaluatable[Callable[[A], Iterable[B]]]
        The function to apply to each element of the iterable. This can be
        an Evaluatable that returns a function, or a constant function.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.flatmap(lambda x: [x, x + 1]) >> list)({'A': [1, 2, 3, 4]})
    [1, 2, 2, 3, 3, 4, 4, 5]
    """
    return PipelineStep(
        map(func) + itertools.chain.from_iterable,
        f"flatmap({func!r})",
    )


def map_items(
    func: MaybeEvaluatable[Callable[[K1, V1], Tuple[K2, V2]]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K2, V2]]:
    """Create a pipeline step that maps a function over the items of a mapping.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[K1, V1], Tuple[K2, V2]]]
        The function to apply to each item of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K2, V2]]
        A pipeline step that applies the function to the items of the mapping.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.map_items(lambda k, v: (k + 1, v + 1)))({'A': {1: 2, 3: 4}})
    mappingproxy({2: 3, 4: 5})
    """
    return PipelineStep(
        (
            Pipeline()  # type: ignore
            + (lambda m: m.items())  # type: ignore
            + map(into(func))  # type: ignore
            + dict
            + MappingProxyType  # type: ignore
        ),
        f"map_items({func!r})",
    )


def map_keys(
    func: MaybeEvaluatable[Callable[[K1], K2]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K2, V1]]:
    """Create a pipeline step that maps a function over the keys of a mapping.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[K1], K2]]
        The function to apply to each key of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K2, V1]]
        A pipeline step that applies the function to the keys of the mapping.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.map_keys(lambda k: k + 1))({'A': {1: 2, 3: 4}})
    mappingproxy({2: 2, 4: 4})
    """
    return PipelineStep(
        map_items(partial(lambda k, v, f: (f(k), v), f=func)),
        f"map_keys({func!r})",
    )


def map_values(
    func: MaybeEvaluatable[Callable[[V1], V2]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K1, V2]]:
    """Create a pipeline step that maps a function over the values of a mapping.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[V1], V2]]
        The function to apply to each value of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K1, V2]]
        A pipeline step that applies the function to the values of the mapping.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.map_values(lambda v: v + 1))({'A': {1: 2, 3: 4}})
    mappingproxy({1: 3, 3: 5})
    """
    return PipelineStep(
        map_items(partial(lambda k, v, f: (k, f(v)), f=func)),
        f"map_values({func!r})",
    )


def filter_items(
    func: MaybeEvaluatable[Callable[[K1, V1], bool]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]:
    """Create a pipeline step that filters items from a mapping using a predicate function.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[K1, V1], bool]]
        The predicate function to apply to each item of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]
        A pipeline step that filters the items of the mapping using the predicate function.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.filter_items(lambda k, v: k % 2 == 0))({'A': {1: 2, 3: 4}})
    mappingproxy({3: 4})
    """
    return PipelineStep(
        (
            Pipeline()  # type: ignore
            + (lambda m: m.items())  # type: ignore
            + filter(into(func))  # type: ignore
            + dict
            + MappingProxyType  # type: ignore
        ),
        f"filter_items({func!r})",
    )


def filter_keys(
    func: MaybeEvaluatable[Callable[[K1], bool]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]:
    """Create a pipeline step that filters keys from a mapping using a predicate function.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[K1], bool]]
        The predicate function to apply to each key of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]
        A pipeline step that filters the keys of the mapping using the predicate function.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.filter_keys(lambda k: k % 2 == 0))({'A': {1: 2, 3: 4}})
    mappingproxy({1: 2})
    """
    return PipelineStep(
        filter_items(partial(lambda k, v, f: f(k), f=func)),
        f"filter_keys({func!r})",
    )


def filter_values(
    func: MaybeEvaluatable[Callable[[V1], bool]],
) -> PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]:
    """Create a pipeline step that filters values from a mapping using a predicate function.

    Arguments
    ---------
    func : MaybeEvaluatable[Callable[[V1], bool]]
        The predicate function to apply to each value of the mapping. This can be an
        Evaluatable that returns a function, or a constant function.

    Returns
    -------
    PipelineStep[Mapping[K1, V1], Mapping[K1, V1]]
        A pipeline step that filters the values of the mapping using the predicate function.

    Example Usage
    -------------
    >>> from labrea import Option
    >>> import labrea.functions as F
    >>>
    >>> (Option('A') >> F.filter_values(lambda v: v % 2 == 0))({'A': {1: 2, 3: 4}})
    mappingproxy({1: 2})
    """
    return PipelineStep(
        filter_items(partial(lambda k, v, f: f(v), f=func)),
        f"filter_values({func!r})",
    )



from typing import List
import uuid

import pytest

from confectioner.templating import set_dotted_key

from labrea.application import FunctionApplication
from labrea.cache import cached, NoCache
from labrea.option import Option
import labrea.cache


def test_cached():
    uuid4 = FunctionApplication(uuid.uuid4)
    cached_uuid4 = cached(uuid4)

    assert uuid4() != uuid4()
    assert cached_uuid4() == cached_uuid4()


def test_nocache():
    uuid4 = FunctionApplication(uuid.uuid4)
    cached_uuid4 = cached(uuid4, NoCache())

    assert uuid4() != uuid4()
    assert cached_uuid4() != cached_uuid4()


def test_cached_decorator():
    @cached
    @FunctionApplication.lift
    def uuid4_list(n: int = Option('N')) -> List[uuid.UUID]:
        return [uuid.uuid4() for _ in range(n)]

    assert uuid4_list({'N': 3}) == uuid4_list({'N': 3})
    assert uuid4_list({'N': 3}) != uuid4_list({'N': 4})

    @cached(NoCache())
    @FunctionApplication.lift
    def uuid4_list(n: int = Option('N')) -> List[uuid.UUID]:
        return [uuid.uuid4() for _ in range(n)]


@pytest.mark.parametrize('method', ['ctx', 'LABREA.CACHE.DISABLED', 'LABREA.CACHE.DISABLE'])
def test_disable_cache(method):
    uuid4 = cached(FunctionApplication(uuid.uuid4))

    a = uuid4()
    b = uuid4()

    if method == 'ctx':
        with labrea.cache.disabled():
            c = uuid4()
            d = uuid4()
    else:
        options = {}
        set_dotted_key(method, True, options)
        c = uuid4(options)
        d = uuid4(options)

    e = uuid4()

    assert a == b == e
    assert a != c and c != d



from labrea.coalesce import Coalesce
from labrea.option import Option
from labrea.exceptions import KeyNotFoundError
import pytest


def test_coalesce():
    a = Option('A')
    b = Option('B')

    c = Coalesce(a, b)

    assert c({'A': 'Hello'}) == 'Hello'
    assert c({'B': 'World!'}) == 'World!'
    assert c({'A': 'Hello', 'B': 'World!'}) == 'Hello'
    assert c({'B': 'World!', 'C': 'foo'}) == 'World!'

    with pytest.raises(KeyNotFoundError):
        c({})

    with pytest.raises(KeyNotFoundError):
        c.validate({})

    with pytest.raises(KeyNotFoundError):
        c.keys({})

    assert c.explain() == {'B'}


def test_repr():
    assert repr(Coalesce(Option('A'), Option('B'))) == "Coalesce(Option('A'), Option('B'))"


def test_init():
    with pytest.raises(TypeError):
        Coalesce()



import pytest

from labrea.arguments import Arguments, EvaluatableArgs, EvaluatableKwargs, arguments
from labrea.exceptions import KeyNotFoundError
from labrea.option import Option


def test_arguments():
    assert repr(Arguments(1, 2, 3)) == "Arguments(1, 2, 3)"
    assert repr(Arguments(a=1, b=2, c=3)) == "Arguments(a=1, b=2, c=3)"
    assert repr(Arguments(1, b=2)) == "Arguments(1, b=2)"
    assert repr(Arguments()) == "Arguments()"

    assert Arguments(1, b=2) == Arguments(1, b=2)
    assert Arguments(1, b=2) != Arguments(1, b=3)
    assert Arguments(1, b=2) != object()



def test_evaluatable_args():
    args = EvaluatableArgs(Option('A'), Option('B'), Option('C'))

    assert args.evaluate({'A': 1, 'B': 2, 'C': 3}) == (1, 2, 3)
    args.validate({'A': 1, 'B': 2, 'C': 3})
    with pytest.raises(KeyNotFoundError):
        args.validate({'A': 1, 'B': 2})
    assert args.keys({'A': 1, 'B': 2, 'C': 3}) == {'A', 'B', 'C'}
    with pytest.raises(KeyNotFoundError):
        args.keys({'A': 1, 'B': 2})
    assert args.explain() == {'A', 'B', 'C'}

    assert repr(args) == "EvaluatableArgs(Option('A'), Option('B'), Option('C'))"


def test_evaluatable_kwargs():
    kwargs = EvaluatableKwargs(a=Option('A'), b=Option('B'), c=Option('C'))

    assert kwargs.evaluate({'A': 1, 'B': 2, 'C': 3}) == {'a': 1, 'b': 2, 'c': 3}
    kwargs.validate({'A': 1, 'B': 2, 'C': 3})
    with pytest.raises(KeyNotFoundError):
        kwargs.validate({'A': 1, 'B': 2})
    assert kwargs.keys({'A': 1, 'B': 2, 'C': 3}) == {'A', 'B', 'C'}
    with pytest.raises(KeyNotFoundError):
        kwargs.keys({'A': 1, 'B': 2})
    assert kwargs.explain() == {'A', 'B', 'C'}

    assert repr(kwargs) == "EvaluatableKwargs(a=Option('A'), b=Option('B'), c=Option('C'))"


def test_evaluatable_arguments():
    args = arguments(Option('A'), b=Option('B'))

    assert args.evaluate({'A': 1, 'B': 2}) == Arguments(1, b=2)
    args.validate({'A': 1, 'B': 2})
    with pytest.raises(KeyNotFoundError):
        args.validate({'A': 1})
    assert args.keys({'A': 1, 'B': 2}) == {'A', 'B'}
    with pytest.raises(KeyNotFoundError):
        args.keys({'A': 1})
    assert args.explain() == {'A', 'B'}

    assert repr(args) == "EvaluatableParameters(Option('A'), b=Option('B'))"



import pytest

from labrea.interface import interface, implements
from labrea.option import Option
from labrea.dataset import dataset, abstractdataset
from labrea.exceptions import EvaluationError


@interface('DISPATCH')
class MyInterface:
    a: str

    @staticmethod
    @abstractdataset
    def b() -> str:
        pass

    def c() -> str:
        return 'c'

    @staticmethod
    @dataset
    def d() -> str:
        return 'd'

    e: str = Option('E', 'e')


@interface('DISPATCH_2')
class MyInterface2:
    f = 'f'


def test_no_implementation():
    with pytest.raises(EvaluationError):
        MyInterface.a()

    with pytest.raises(EvaluationError):
        MyInterface.b()

    assert MyInterface.c() == 'c'
    assert MyInterface.d() == 'd'
    assert MyInterface.e() == 'e'


def test_good_implementation():
    @MyInterface.implementation('GOOD')
    class Good:
        a = 'a'
        b = 'b'

    assert MyInterface.a({'DISPATCH': 'GOOD'}) == 'a'
    assert MyInterface.b({'DISPATCH': 'GOOD'}) == 'b'
    assert MyInterface.c({'DISPATCH': 'GOOD'}) == 'c'
    assert MyInterface.d({'DISPATCH': 'GOOD'}) == 'd'
    assert MyInterface.e({'DISPATCH': 'GOOD'}) == 'e'


def test_bad_implementation():
    with pytest.raises(TypeError):
        @MyInterface.implementation('BAD')
        class Bad:
            pass

    with pytest.raises(TypeError):
        @MyInterface.implementation('BAD')
        class Bad:
            a = 'a'

    with pytest.raises(TypeError):
        @MyInterface.implementation('BAD')
        class Bad:
            b = 'b'

    with pytest.raises(TypeError):
        @MyInterface.implementation('BAD')
        class Bad:
            a = 'a'
            b = 'b'
            f = 'f'


def test_override():
    @MyInterface.implementation('OVERRIDE')
    class Good:
        a = 'A'
        b = 'B'

        def c():
            return 'C'

        @staticmethod
        @dataset
        def d():
            return 'D'

        e = Option('E', 'E')

    assert MyInterface.a({'DISPATCH': 'OVERRIDE'}) == 'A'
    assert MyInterface.b({'DISPATCH': 'OVERRIDE'}) == 'B'
    assert MyInterface.c({'DISPATCH': 'OVERRIDE'}) == 'C'
    assert MyInterface.d({'DISPATCH': 'OVERRIDE'}) == 'D'
    assert MyInterface.e({'DISPATCH': 'OVERRIDE'}) == 'E'


def test_multi_implementation():
    @implements(MyInterface, MyInterface2, alias='MULTI')
    class Multi:
        a = 'a'
        b = 'b'

    assert MyInterface.a({'DISPATCH': 'MULTI'}) == 'a'
    assert MyInterface.b({'DISPATCH': 'MULTI'}) == 'b'
    assert MyInterface.c({'DISPATCH': 'MULTI'}) == 'c'
    assert MyInterface.d({'DISPATCH': 'MULTI'}) == 'd'
    assert MyInterface.e({'DISPATCH_2': 'MULTI'}) == 'e'
    assert MyInterface2.f({'DISPATCH_2': 'MULTI'}) == 'f'


def test_missing_args():
    with pytest.raises(TypeError):
        @interface
        class A:
            pass

    with pytest.raises(TypeError):
        @MyInterface.implementation
        class A:
            pass


def test_invalid_member():
    with pytest.raises(AttributeError):
        @interface('A')
        class A:
            implementation = 'implementation'


def test_repr():
    @MyInterface.implementation('A')
    class A:
        a: str = 'a'
        b: str = 'b'

    assert repr(MyInterface) == '<Interface test_interface.MyInterface>'
    assert repr(A) == '<Implementation test_interface.A of MyInterface>'



from labrea.collections import (
    evaluatable_dict,
    evaluatable_list,
    evaluatable_set,
    evaluatable_tuple,
)
from labrea.option import Option


def test_dict():
    assert evaluatable_dict({'a': Option('A'), 'b': 2}).evaluate({'A': 1}) == {'a': 1, 'b': 2}


def test_list():
    assert evaluatable_list(Option('A'), 2).evaluate({'A': 1}) == [1, 2]


def test_set():
    assert evaluatable_set(Option('A'), 2).evaluate({'A': 1}) == {1, 2}


def test_tuple():
    assert evaluatable_tuple(Option('A'), 2).evaluate({'A': 1}) == (1, 2)


import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import functools
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    List,
    Optional,
    Set,
    TypeVar,
    Union,
    overload,
)

from confectioner import mix

from ._missing import MISSING, MaybeMissing
from .application import FunctionApplication
from .cache import Cache, MemoryCache, NoCache, cached
from .computation import CallbackEffect, ChainedEffect, Computation, Effect
from .logging import Logged
from .option import Option, WithDefaultOptions, WithOptions
from .overload import Overloaded
from .types import Evaluatable, MaybeEvaluatable, Options, Value

A = TypeVar("A", covariant=True)
P = ParamSpec("P")
Callback = MaybeEvaluatable[Callable[[A], None]]


class Dataset(Evaluatable[A]):
    """A class representing a dataset.

    Datasets are the building blocks of Labrea programs. They represent the result of performing
    some calculation over a set of inputs. Datasets can be composed together to create complex
    call graphs that represent the flow of data through a program. Datasets are parameterized
    using :code:`Options`, which allow for user input.

    Datasets are not created directly; instead, they are created using the :code:`dataset`
    decorator. This decorator takes a function that defines the dataset and returns a new
    :code:`Dataset` object. When a dataset is evaluated, the default values to each argument
    are evaluated using the provided options. The dataset is then evaluated using the
    evaluated arguments.


    Example Usage
    -------------
    >>> @dataset
    ... def a_squared(a: float = Option('A')) -> float:
    ...     return a ** 2
    >>>
    >>> @dataset
    ... def b_squared(b: float = Option('B')) -> float:
    ...     return b ** 2
    >>>
    >>> @dataset
    ... def hypotenuse(a2: float = a_squared, b2: float = b_squared) -> float:
    ...     return (a2 + b2) ** 0.5
    >>>
    >>> hypotenuse({'A': 3, 'B': 4})
    5.0
    """

    overloads: Overloaded[A]
    effects: List[Effect[A]]
    cache: Cache[A]
    options: Options
    default_options: Options
    _effects_disabled: bool

    def __init__(
        self,
        overloads: Overloaded[A],
        effects: List[Effect[A]],
        cache: Cache[A],
        options: Options,
        default_options: Options,
    ):
        self.overloads = overloads
        self.effects = effects.copy()
        self.cache = cache
        self.options = options
        self.default_options = default_options
        self._effects_disabled = False

    @property
    def _composed(self) -> Evaluatable[A]:
        computation = Computation(
            self.overloads,
            ChainedEffect(*self.effects),
        )
        return WithDefaultOptions(
            WithOptions(
                cached(
                    Logged(
                        self.overloads if self._effects_disabled else computation,
                        level=logging.INFO,
                        name=self.__module__,
                        msg=f"Labrea: Evaluating {self!r}",
                    ),
                    self.cache,
                ),
                self.options,
            ),
            self.default_options,
        )

    def evaluate(self, options: Options) -> A:
        """Evaluates the dataset using the provided options."""
        return self._composed.evaluate(options)

    def validate(self, options: Options) -> None:
        """Validates the dataset using the provided options."""
        self._composed.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Returns the option keys required by the dataset."""
        return self._composed.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Returns the option keys required by the dataset."""
        return self._composed.explain(options)

    def __repr__(self) -> str:
        kind = "AbstractDataset" if self.is_abstract else "Dataset"
        if hasattr(self, "__qualname__"):
            return f"<{kind} {self.__qualname__}>"
        if hasattr(self, "__name__"):
            return f"<{kind} {self.__name__}>"

        return (  # pragma: no cover
            f"Dataset("
            f"{self.overloads!r}, "
            f"{', '.join(map(repr, self.effects))}, "
            f"{self.cache!r})"
        )

    def overload(
        self,
        alias: Union[Hashable, List[Hashable]],
    ) -> Callable[[Callable[P, A]], FunctionApplication[P, A]]:
        """Overloads the dataset with a new implementation. Used as a decorator.

        Overloading a dataset allows you to provide a new implementation for the dataset.
        Overloads are selected based on the dispatch value provided to the dataset. If no
        dispatch value is provided, the default implementation is used. Datasets can also
        be abstract, in which no default implementation is provided, and an error is raised
        if no overload can be found.

        Datasets must be created with a dispath argument in order to use overloads. If you
        are trying to overload a dataset that you do not own (i.e. from a third-party library),
        you can use my_dataset.set_dispatch(dispatch) to set the dispatch manually. The
        dispatch can be any Evaluatable object that returns a hashable value, or a string
        representing the Option key to use.

        Arguments
        ---------
        alias : Union[Hashable, List[Hashable]]
            The name or names to use for the overload.

        Returns
        -------
        FunctionApplication[P, A]
            The Evaluatable object representing the new implementation.


        Example Usage
        -------------
        >>> @dataset(dispatch='INPUT.SOURCE'))
        ... def input_data(path: str = Option('INPUT.PATH')) -> list[str]:
        ...     with open(path, 'r') as f:
        ...         return f.read()
        >>>
        >>> @input_data.overload('MOCK')
        ... def mock_input_data() -> list[str]:
        ...     return ['Mock', 'Data']
        >>>
        >>> input_data({'INPUT': {'PATH': 'data.txt'}})
        ['Data', 'From', 'File']
        >>> input_data({'INPUT': {'SOURCE': 'MOCK'}})
        ['Mock', 'Data']
        """
        if self.overloads.dispatch == Value(MISSING):
            raise ValueError(
                "Cannot add overloads to a dataset without a dispatch. Add a dispatch by using "
                "@dataset(dispatch=...) in the dataset definition. If you are trying to overload "
                "a dataset that you do not own (i.e. from a third-party library), you can use "
                "my_dataset.set_dispatch(dispatch) to set the dispatch manually."
            )

        if not isinstance(alias, list):
            alias = [alias]

        def decorator(func: Callable[P, A]) -> FunctionApplication[P, A]:
            overload_ = FunctionApplication.lift(func)
            for key in alias:
                self.register(key, overload_)
            return overload_

        return decorator

    def register(self, key: Hashable, value: Evaluatable[A]) -> None:
        """Registers a new overload for the dataset.

        Registers a new overload for the dataset. Can be used if you want an overload to be
        a pre-existing Evaluatable object, like an Option.

        Arguments
        ---------
        key : Hashable
            The name to use for the overload.
        value : Evaluatable[A]
            The Evaluatable object representing the new implementation.


        Example Usage
        -------------
        >>> @dataset(dispatch='INPUT.SOURCE'))
        ... def input_data(path: str = Option('INPUT.PATH')) -> list[str]:
        ...     with open(path, 'r') as f:
        ...         return f.read()
        >>>
        >>> input_data.register('MOCK', Value(['Mock', 'Data']))
        """
        self.overloads.register(key, value)

    def set_dispatch(self, dispatch: Evaluatable[Hashable]) -> None:
        """Sets the dispatch value for the dataset.

        This is a stateful operation that changes the dispatch value for the dataset. This
        is intended for use with third-party datasets that do not have a dispatch value set,
        but you want to add an overload to.

        Arguments
        ---------
        dispatch : Evaluatable[Hashable]
            The dispatch value to use for the dataset.
        """

        self.overloads = Overloaded(
            dispatch,
            self.overloads.lookup.copy(),
            default=self.overloads.default,
        )

    def set_cache(self, cache: Union[Cache[A], Callable[..., Cache[A]]]) -> None:
        """Sets the cache for the dataset.

        Sets the cache for the dataset. The cache can be any object that implements the Cache
        interface, or a callable that returns a Cache object. This is a stateful operation that
        changes the cache for the dataset. This is intended for use with third-party datasets
        that do not have a cache set, but you want to add a cache to.

        Arguments
        ---------
        cache : Union[Cache[A], Callable[..., Cache[A]]]
            The cache to use for the dataset.
        """
        cache = cache if isinstance(cache, Cache) else cache()
        if not isinstance(cache, Cache):
            raise TypeError(f"Invalid cache: {cache}")

        self.cache = cache

    def add_effects(self, *effects: Union[Effect[A], Callback[A]]) -> None:
        """Adds effects to the dataset.

        Adds effects to the dataset. Effects are applied to the dataset when it is evaluated.
        This is a stateful operation that changes the effects for the dataset. This is intended
        for use with third-party datasets that you would like to add effects to (i.e. logging).

        Arguments
        ---------
        effects : Union[Effect[A], Callback[A]]
            The effects to add to the dataset.
        """
        for effect in effects:
            if isinstance(effect, Effect):
                self.effects.append(effect)
            else:
                self.effects.append(CallbackEffect(effect))

    add_effect = add_effects

    def disable_effects(self) -> None:
        """Disables effects for the dataset.

        Disables effects for the dataset. Effects are not executed when the dataset is evaluated.
        This is a stateful operation that changes the effects for the dataset. This is intended
        for use with third-party datasets that you would like to disable effects for.
        """
        self._effects_disabled = True

    def enable_effects(self) -> None:
        """Enables effects for the dataset.

        Enables effects for the dataset. Effects are executed when the dataset is evaluated.
        This is a stateful operation that changes the effects for the dataset. This is intended
        for use with third-party datasets that you would like to enable effects for.
        """
        self._effects_disabled = False

    def with_options(self, options: Options) -> "Dataset[A]":
        """Returns a new dataset with the provided options pre-set.

        The provided options are pre-set, and cannot be overridden by the user. This
        can simplify the creation of multiple datasets that have the same inputs and
        calculations.

        Arguments
        ---------
        options : Options
            The options to pre-set for the dataset.

        Returns
        -------
        Dataset[A]
            A new dataset with the provided options pre-set.


        Example Usage
        -------------
        >>> @dataset
        ... def a_to_power(a: float = Option('A'), power: float = Option('POWER')) -> float:
        ...     return a ** power
        >>>
        >>> a_squared = a_to_power.with_options({'POWER': 2})
        """
        return Dataset(
            self.overloads,
            self.effects,
            self.cache,
            mix(self.options, options),  # type: ignore
            self.default_options,
        )

    def with_default_options(self, options: Options) -> "Dataset[A]":
        """Returns a new dataset with the provided options as default.

        The provided options are set as the default options for the dataset. These options
        can be overridden by the user. This can simplify the creation of multiple datasets
        that have the same inputs and calculations.

        Arguments
        ---------
        options : Options
            The options to set as the default for the dataset.

        Returns
        -------
        Dataset[A]
            A new dataset with the provided options as default.
        """
        return Dataset(
            self.overloads,
            self.effects,
            self.cache,
            self.options,
            mix(self.default_options, options),  # type: ignore
        )

    @property
    def default(self) -> MaybeMissing[Evaluatable[A]]:
        """The default implementation of the dataset."""
        return self.overloads.default

    @property
    def is_abstract(self) -> bool:
        """Whether the dataset is abstract (i.e. has no default implementation)."""
        return self.overloads.default is MISSING


class DatasetFactory(Generic[A]):
    effects: List[Effect[A]]
    cache: Union[Cache[A], Callable[..., Cache[A]], None]
    dispatch: Evaluatable[Hashable]
    defaults: Dict[str, Evaluatable[Any]]
    options: Options
    default_options: Options
    abstract: bool

    def __init__(
        self,
        effects: Optional[List[Effect[A]]] = None,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = None,
        dispatch: Union[Evaluatable[Hashable], str, None] = None,
        defaults: Optional[Dict[str, Any]] = None,
        abstract: bool = False,
        options: Optional[Options] = None,
        default_options: Optional[Options] = None,
    ):
        self.effects = effects or []
        self.cache = cache
        self.defaults = {
            key: Evaluatable.ensure(val) for key, val in (defaults or {}).items()
        }
        self.abstract = abstract
        self.options = options or {}
        self.default_options = default_options or {}

        if dispatch is None:
            self.dispatch = Value(MISSING)
        elif isinstance(dispatch, str):
            self.dispatch = Option(dispatch)
        else:
            self.dispatch = dispatch

    @overload
    def __call__(
        self,
        definition: Callable[..., A],
        /,
        *,
        effects: Optional[List[Union[Effect[A], Callback[A]]]] = ...,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = ...,
        dispatch: Optional[Union[Evaluatable[Hashable], str]] = ...,
        defaults: Optional[Dict[str, Any]] = ...,
        abstract: Optional[bool] = ...,
        options: Optional[Options] = ...,
        default_options: Optional[Options] = ...,
    ) -> Dataset[A]: ...  # pragma: no cover

    @overload
    def __call__(
        self,
        /,
        *,
        effects: Optional[List[Union[Effect[A], Callback[A]]]] = ...,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = ...,
        dispatch: Optional[Union[Evaluatable[Hashable], str]] = ...,
        defaults: Optional[Dict[str, Any]] = ...,
        abstract: Optional[bool] = ...,
        options: Optional[Options] = ...,
        default_options: Optional[Options] = ...,
    ) -> "DatasetFactory[A]": ...  # pragma: no cover

    @overload
    def __call__(
        self,
        definition: None,
        /,
        *,
        effects: Optional[List[Union[Effect[A,], Callback[A]]]] = ...,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = ...,
        dispatch: Optional[Union[Evaluatable[Hashable], str]] = ...,
        defaults: Optional[Dict[str, Any]] = ...,
        abstract: Optional[bool] = ...,
        options: Optional[Options] = ...,
        default_options: Optional[Options] = ...,
    ) -> "DatasetFactory[A]": ...  # pragma: no cover

    def __call__(
        self,
        definition: Optional[Callable[..., A]] = None,
        /,
        *,
        effects: Optional[List[Union[Effect[A], Callback[A]]]] = None,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = None,
        dispatch: Optional[Union[Evaluatable[Hashable], str]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        abstract: Optional[bool] = None,
        options: Optional[Options] = None,
        default_options: Optional[Options] = None,
    ) -> Union["DatasetFactory[A]", Dataset[A]]:
        _effects = [
            effect if isinstance(effect, Effect) else CallbackEffect(effect)
            for effect in (effects or [])
        ]
        factory = self.update(
            effects=_effects,
            cache=cache,
            dispatch=dispatch,
            defaults=defaults,
            abstract=abstract,
            options=options,
            default_options=default_options,
        )

        if definition is not None:
            return factory.wrap(definition)

        return factory

    def wrap(self, definition: Callable[..., A]) -> Dataset[A]:
        overloads: Overloaded[A]
        if not self.abstract:
            lifted: Evaluatable[A] = (
                FunctionApplication.lift(definition, **self.defaults)
                if not isinstance(definition, Evaluatable)
                else definition
            )
            overloads = Overloaded(self.dispatch, {}, lifted)
        else:
            overloads = Overloaded(self.dispatch, {})

        cache: Cache
        if self.cache is None:
            cache = MemoryCache()
        elif callable(self.cache):
            cache = self.cache()
        elif isinstance(self.cache, Cache):
            cache = self.cache
        else:
            raise TypeError(f"Invalid cache: {self.cache}")

        _dataset = Dataset(
            overloads,
            effects=self.effects,
            cache=cache,
            options=self.options,
            default_options=self.default_options,
        )

        functools.update_wrapper(_dataset, definition, updated=())

        return _dataset

    def where(self, **defaults: Any) -> "DatasetFactory":
        return self.update(defaults=defaults)

    def update(
        self,
        effects: Optional[List[Effect[A]]] = None,
        cache: Union[Cache[A], Callable[..., Cache[A]], None] = None,
        dispatch: Optional[Union[Evaluatable[Hashable], str]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        abstract: Optional[bool] = None,
        options: Optional[Options] = None,
        default_options: Optional[Options] = None,
    ) -> "DatasetFactory":
        return DatasetFactory(
            effects=[*self.effects, *(effects or [])],
            cache=cache or self.cache,
            dispatch=dispatch or self.dispatch,
            defaults={**self.defaults, **(defaults or {})},
            abstract=abstract if abstract is not None else self.abstract,
            options=options or self.options,
            default_options=default_options or self.default_options,
        )

    @property
    def nocache(self) -> "DatasetFactory":
        return self.update(cache=NoCache())


dataset: DatasetFactory = DatasetFactory()
abstractdataset: DatasetFactory = dataset(abstract=True)



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, Optional, Set, TypeVar, Union, overload

from . import runtime
from .option import Option
from .types import Evaluatable, Options

A = TypeVar("A")


class CacheFailure(Exception):
    """An exception raised when a cache operation fails."""

    evaluatable: Evaluatable
    options: Options
    cache: "Cache"


class CacheGetFailure(CacheFailure):
    """An exception raised when a cache get operation fails."""

    def __init__(
        self, evaluatable: Evaluatable, options: Options, cache: "Cache"
    ) -> None:
        self.evaluatable = evaluatable
        self.options = options
        self.cache = cache

        super().__init__(f"Failed to retrieve {evaluatable} from cache {cache}")


class CacheSetFailure(CacheFailure):
    """An exception raised when a cache set operation fails."""

    evaluatable: Evaluatable
    options: Options
    cache: "Cache"
    value: Any

    def __init__(
        self, evaluatable: Evaluatable, options: Options, cache: "Cache", value: Any
    ) -> None:
        self.evaluatable = evaluatable
        self.options = options
        self.cache = cache
        self.value = value

        super().__init__(f"Failed to add value {value} from {evaluatable} to {cache}")


class CacheExistsFailure(CacheFailure):
    """An exception raised when a cache exists operation fails."""

    evaluatable: Evaluatable
    options: Options
    cache: "Cache"

    def __init__(
        self, evaluatable: Evaluatable, options: Options, cache: "Cache"
    ) -> None:
        self.evaluatable = evaluatable
        self.options = options
        self.cache = cache

        super().__init__(f"Failed to check if {evaluatable} exists in cache {cache}")


class Cache(Generic[A], ABC):
    """A class representing a cache of values that can be evaluated."""

    @abstractmethod
    def get(self, evaluatable: Evaluatable[A], options: Options) -> A:
        """Get a key from the cache.

        Arguments
        ---------
        evaluatable: Evaluatable
            The evaluatable to get from the cache.
        options : Options
            The options dictionary to evaluate against.

        Returns
        -------
        A
            The value from the cache.

        Raises
        ------
        CacheGetFailure
            If the key cannot be retrieved from the cache.
        """
        raise NotImplementedError  # pragma: nocover

    @abstractmethod
    def set(self, evaluatable: Evaluatable[A], options: Options, value: A) -> None:
        """Set a key in the cache.

        Arguments
        ---------
        evaluatable: Evaluatable
            The evaluatable to set in the cache.
        options : Options
            The options dictionary to evaluate against.
        value : A
            The value to set in the cache.

        Raises
        ------
        CacheSetFailure
            If the key cannot be set in the cache.
        """
        raise NotImplementedError  # pragma: nocover

    def exists(self, evaluatable: Evaluatable[A], options: Options) -> bool:
        """Check if a key exists in the cache.

        Arguments
        ---------
        evaluatable: Evaluatable
            The evaluatable to check in the cache.
        options : Options
            The options dictionary to evaluate against.

        Returns
        -------
        bool
            Whether the value exists in the cache.

        Raises
        ------
        CacheExistsFailure
            If the existence of the key cannot be checked in the cache.
        """
        try:
            self.get(evaluatable, options)
            return True
        except CacheGetFailure:
            return False


class NoCache(Cache[Any]):
    """A class representing a cache that does not store any values."""

    def get(self, evaluatable: Evaluatable, options: Options) -> Any:
        raise CacheGetFailure(evaluatable, options, self)

    def set(self, evaluatable: Evaluatable, options: Options, value: Any) -> None:
        pass


class MemoryCache(Cache[A]):
    """A class representing a cache that stores values in memory."""

    _cache: Dict[bytes, A]

    def __init__(self) -> None:
        self._cache = {}

    def get(self, evaluatable: Evaluatable, options: Options) -> A:
        try:
            return self._cache[evaluatable.fingerprint(options)]
        except KeyError as e:
            raise CacheGetFailure(evaluatable, options, self) from e

    def set(self, evaluatable: Evaluatable, options: Options, value: A) -> None:
        self._cache[evaluatable.fingerprint(options)] = value

    def exists(self, evaluatable: Evaluatable, options: Options) -> bool:
        return evaluatable.fingerprint(options) in self._cache


class CacheSetRequest(runtime.Request[A]):
    """A request to set a value in a cache.

    Arguments
    ---------
    evaluatable : Evaluatable
        The evaluatable that produced the value.
    options : Options
        The options dictionary that was used to evaluate the value.
    value : A
        The value to set in the cache.
    cache : Cache
        The cache to set the value in.
    """

    evaluatable: Evaluatable
    options: Options
    value: A
    cache: Cache[A]

    def __init__(
        self, evaluatable: Evaluatable[A], options: Options, value: A, cache: Cache[A]
    ):
        self.evaluatable = evaluatable
        self.options = options
        self.value = value
        self.cache = cache


class CacheGetRequest(runtime.Request[A]):
    """A request to get a value from a cache.

    Arguments
    ---------
    evaluatable : Evaluatable
        The evaluatable to get from the cache.
    options : Options
        The options dictionary to evaluate against.
    cache : Cache
        The cache to get the value from.
    """

    evaluatable: Evaluatable[A]
    options: Options
    cache: Cache[A]

    def __init__(self, evaluatable: Evaluatable, options: Options, cache: Cache[A]):
        self.evaluatable = evaluatable
        self.options = options
        self.cache = cache


class CacheExistsRequest(runtime.Request[bool]):
    """A request to check if a value exists in a cache.

    Arguments
    ---------
    evaluatable : Evaluatable
        The evaluatable to check in the cache.
    options : Options
        The options dictionary to evaluate against.
    cache : Cache
        The cache to check the value in.
    """

    evaluatable: Evaluatable
    options: Options
    cache: Cache

    def __init__(self, evaluatable: Evaluatable, options: Options, cache: Cache):
        self.evaluatable = evaluatable
        self.options = options
        self.cache = cache


def _cache_disabled(
    request: Union[CacheSetRequest, CacheGetRequest, CacheExistsRequest],
) -> bool:
    return Option("LABREA.CACHE.DISABLED", Option("LABREA.CACHE.DISABLE", False))(
        request.options
    )


@CacheSetRequest.handle
def _set_cache_handler(request: CacheSetRequest[A]) -> A:
    if _cache_disabled(request):
        return _disabled_set_cache_handler(request)

    request.cache.set(request.evaluatable, request.options, request.value)
    try:
        return request.cache.get(request.evaluatable, request.options)
    except CacheGetFailure:
        return request.value


@CacheGetRequest.handle
def _get_cache_handler(request: CacheGetRequest[A]) -> A:
    if _cache_disabled(request):
        return _disabled_get_cache_handler(request)

    return request.cache.get(request.evaluatable, request.options)


@CacheExistsRequest.handle
def _exists_cache_handler(request: CacheExistsRequest) -> bool:
    if _cache_disabled(request):
        return _disabled_exists_cache_handler(request)

    return request.cache.exists(request.evaluatable, request.options)


def _disabled_set_cache_handler(request: CacheSetRequest[A]) -> A:
    return request.value


def _disabled_get_cache_handler(request: CacheGetRequest[A]) -> A:
    raise CacheGetFailure(request.evaluatable, request.options, request.cache)


def _disabled_exists_cache_handler(request: CacheExistsRequest) -> bool:
    return False


def disabled() -> runtime.Runtime:
    return runtime.handle(
        {
            CacheSetRequest: _disabled_set_cache_handler,
            CacheGetRequest: _disabled_get_cache_handler,
            CacheExistsRequest: _disabled_exists_cache_handler,
        }
    )


class Cached(Evaluatable[A]):
    """A class representing an Evaluatable that may be cached.

    When evaluated, this class will check if the value exists in the cache
    before evaluating the underlying Evaluatable. If the value is not in the
    cache, the value will be evaluated and stored in the cache before being
    returned.

    Arguments
    ---------
    evaluatable : Evaluatable
        The evaluatable to cache.
    cache : Cache
        The cache to store the value in.
    """

    evaluatable: Evaluatable[A]
    cache: Cache[A]

    def __init__(self, evaluatable: Evaluatable[A], cache: Cache[A]):
        self.evaluatable = evaluatable
        self.cache = cache

    def evaluate(self, options: Options) -> A:
        """Return the (possibly cached) result of evaluating the evaluatable."""
        if CacheExistsRequest(self.evaluatable, options, self.cache).run():
            try:
                return CacheGetRequest(self.evaluatable, options, self.cache).run()
            except CacheGetFailure:
                pass

        value = self.evaluatable.evaluate(options)

        return CacheSetRequest(self.evaluatable, options, value, self.cache).run()

    def validate(self, options: Options) -> None:
        """If the value is not in the cache, validate the evaluatable."""
        if not CacheExistsRequest(self.evaluatable, options, self.cache).run():
            self.evaluatable.validate(options)

    def keys(self, options: Options) -> Set[str]:
        """Return the keys required to evaluate the evaluatable."""
        return self.evaluatable.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the keys required to evaluate the evaluatable."""
        return self.evaluatable.explain(options)

    def __repr__(self) -> str:
        return f"Cached({self.evaluatable!r}, {self.cache!r})"


@overload
def cached(
    __x: Evaluatable[A], cache: Optional[Cache[A]] = None
) -> Evaluatable[A]: ...  # pragma: nocover


@overload
def cached(
    __x: Cache[A],
) -> Callable[[Evaluatable[A]], Evaluatable[A]]: ...  # pragma: nocover


def cached(
    __x: Union[Cache[A], Evaluatable[A]],
    cache: Optional[Cache[A]] = None,
) -> Union[Callable[[Evaluatable[A]], Evaluatable[A]], Evaluatable[A]]:
    """Return an Evaluatable that caches the result of evaluating the input Evaluatable.

    Can be used as a decorator or a function. If used as a decorator, can be used with or without
    providing a cache object. If no cache object is provided, a MemoryCache will be used.

    Arguments
    ---------
    __x : Union[Cache, Evaluatable]
        The cache object or evaluatable to cache.
    cache : Optional[Cache], optional
        The cache object to use, by default None, in which case a MemoryCache is used.

    Returns
    -------
    Union[Callable[[Evaluatable[A]], Evaluatable[A]], Evaluatable[A]]
        The cached evaluatable, or a function that takes an evaluatable and returns a cached
        evaluatable.


    Example Usage
    -------------
    >>> from labrea import cached, Option
    >>> from labrea.cache import MemoryCache
    >>> from labrea.application import FunctionApplication
    >>> x_cached = cached(Option('X'))
    >>> x_cached_with_cache = cached(Option('X'), MemoryCache())
    >>>
    >>> @cached
    ... @FunctionApplication.lift
    ... def y(x: int = Option('X')) -> int:
    ...     return x
    >>>
    >>> @cached(MemoryCache())
    ... @FunctionApplication.lift
    ... def z(x: int = Option('X')) -> int:
    ...     return x
    """
    if isinstance(__x, Cache):
        return lambda evaluatable: cached(evaluatable, __x)
    else:
        cache = cache or MemoryCache()
        return Cached(__x, cache)



import pytest

from labrea.types import Evaluatable, Value
from labrea.exceptions import EvaluationError, KeyNotFoundError
from labrea.option import Option


def test_value():
    value = Value(42)
    assert value.evaluate({}) == value() == 42
    assert value.validate({}) is None
    assert value.keys({}) == set()
    assert value.explain() == value.explain({}) == set()
    assert repr(value) == "Value(42)"
    assert value == value
    assert value != Value(43)

    class Uncopyable:
        def __deepcopy__(self, memo):
            raise NotImplementedError('Cannot copy this.')

    uncopyable = Uncopyable()

    assert Value(uncopyable).evaluate({}) is uncopyable


def test_unit():
    assert Evaluatable.unit(1)() == 1


@pytest.mark.parametrize(
    'wrapper,method',
    [
        (wrapper, method)
        for wrapper in (Value, lambda x: x)
        for method in ('apply', 'rshift')
    ]
)
def test_apply(wrapper, method):
    value = Value(42)

    def incr(x):
        return x + 1

    if method == 'apply':
        apply = value.apply(wrapper(incr))
    else:
        apply = value >> wrapper(incr)

    assert apply.evaluate({}) == 43
    assert apply.validate({}) is None
    assert apply.keys({}) == set()
    assert apply.explain() == set()
    assert repr(apply) == f"Value(42).apply(Value({repr(incr)}))"


def test_bind():
    value = Value(42)

    def incr(x):
        return Value(x + 1)

    bind = value.bind(incr)
    assert bind.evaluate({}) == 43
    assert bind.validate({}) is None
    assert bind.keys({}) == set()
    assert bind.explain() == set()
    assert repr(bind) == f"Value(42).bind({repr(incr)})"


def test_type_error():
    with pytest.raises(TypeError):
        Value(42).bind(42)

    with pytest.raises(TypeError):
        Value(42).apply(42)


def test_error_str():
    assert str(EvaluationError('message', Value(1))) == "Originating in Value(1) | message"
    assert str(KeyNotFoundError('key', Value(1))) == "Originating in Value(1) | Key 'key' not found"


def test_fingerprint():
    value = Value(42)
    option = Option('A')

    assert value.fingerprint({}) == value.fingerprint({'A': 1})
    assert value.fingerprint({'A': 1}) != option.fingerprint({'A': 1})
    assert option.fingerprint({'A': 1}) == option.fingerprint({'A': 1})
    assert option.fingerprint({'A': 1}) != option.fingerprint({'A': 2})
    assert option.fingerprint({'A': 1}) == option.fingerprint({'A': 1, 'V': 2})


def test_result():
    option = Option('A')
    assert option.result is option



import pytest

from labrea import datasetclass, Option
from labrea.exceptions import KeyNotFoundError


@datasetclass
class X:
    a: int = Option('A')
    b: bool = True
    c: str = Option('C').apply(str)


@datasetclass
class X2:
    a: int = Option('A')
    b: bool = True
    c: str = Option('C').apply(str)


def test_evaluate():
    x = X({'A': 1, 'C': 3})

    assert x.a == 1
    assert x.b is True
    assert x.c == '3'

    with pytest.raises(KeyNotFoundError) as exc_info:
        X({'A': 1})
        assert exc_info.value.key == 'C'


def test_validate():
    X.validate({'A': 1, 'C': 3})

    with pytest.raises(KeyNotFoundError) as exc_info:
        X.validate({'A': 1})
        assert exc_info.value.key == 'C'


def test_keys():
    assert X.keys({'A': 1, 'C': 3}) == {'A', 'C'}

    with pytest.raises(KeyNotFoundError) as exc_info:
        X.keys({'A': 1})
        assert exc_info.value.key == 'C'


def test_explain():
    assert X.explain() == X.explain({'A': 1, 'C': 3}) == {'A', 'C'}


def test_repr():
    assert repr(X) == '<DatasetClass X>'
    opts = {'A': 1, 'C': 3}
    assert repr(X(opts)) == f'X({opts!r})'


def test_missing_default():
    with pytest.raises(ValueError):
        @datasetclass
        class Y:
            a: int = Option('A')
            b: bool
            c: str = Option('C').apply(str)


def test_eq():
    assert X({'A': 1, 'C': 3}) == X({'A': 1, 'C': 3})
    assert X({'A': 1, 'C': 3}) != X({'A': 1, 'C': 4})
    assert X({'A': 1, 'C': 3}) != X2({'A': 1, 'C': 3})


def test_inheritance():
    @datasetclass
    class Y(X):
        d: float = Option('D')

    y = Y({'A': 1, 'C': 3, 'D': 4.0})
    assert y.a == 1
    assert y.b is True
    assert y.c == '3'
    assert y.d == 4.0



from labrea import Option
from labrea.application import FunctionApplication, PartialApplication
from labrea.exceptions import KeyNotFoundError
import pytest


def add(x: float, y: float) -> float:
    return x + y


def test_application():
    a = Option('A')
    b = Option('B')

    app = FunctionApplication(add, a, b)

    assert app.evaluate({'A': 1, 'B': 2}) == 3
    with pytest.raises(KeyNotFoundError):
        app.evaluate({'A': 1})

    app.validate({'A': 1, 'B': 2})
    with pytest.raises(KeyNotFoundError):
        app.validate({'A': 1})

    assert app.keys({'A': 1, 'B': 2}) == {'A', 'B'}
    with pytest.raises(KeyNotFoundError):
        app.keys({'A': 1})

    assert app.explain() == {'A', 'B'}

    assert repr(app) == "FunctionApplication(add, Option('A'), Option('B'))"


def test_lift():
    def good(a: float = Option('A'), b: float = Option('B')) -> float:
        return a + b

    def bad(a: float, b: float) -> float:
        return a + b

    assert FunctionApplication.lift(good).evaluate({'A': 1, 'B': 2}) == 3

    with pytest.raises(TypeError):
        FunctionApplication.lift(bad)

    assert FunctionApplication.lift(bad, a=Option('A'), b=Option('B')).evaluate({'A': 1, 'B': 2}) == 3

    @FunctionApplication.lift
    def good_deco(a: float = Option('A'), b: float = Option('B')) -> float:
        return a + b

    assert good_deco.evaluate({'A': 1, 'B': 2}) == 3

    @FunctionApplication.lift(a=Option('A'), b=Option('B'))
    def good_deco_kwargs(a: float, b: float = Option('X')) -> float:
        return a + b

    assert good_deco_kwargs.evaluate({'A': 1, 'B': 2}) == 3


def test_partial_application():
    b = Option('B')

    app = PartialApplication(add, b)

    assert app.evaluate({'B': 2})(1) == 3
    with pytest.raises(KeyNotFoundError):
        app.evaluate({})

    app.validate({'B': 2})
    with pytest.raises(KeyNotFoundError):
        app.validate({})

    assert app.keys({'B': 2}) == {'B'}
    with pytest.raises(KeyNotFoundError):
        app.keys({})

    assert app.explain() == {'B'}

    assert repr(app) == "PartialApplication(add, Option('B'))"


def test_lift_partial():
    def some(a: float, b: float = Option('B')) -> float:
        return a + b

    def none(a: float, b: float) -> float:
        return a + b

    def all(a: float = Option('A'), b: float = Option('B')) -> float:
        return a + b

    assert PartialApplication.lift(some).evaluate({'B': 2})(1) == 3
    assert PartialApplication.lift(some).evaluate({'B': 2})(a=1, b=3) == 4
    assert PartialApplication.lift(some, b=Option('C')).evaluate({'C': 4})(1) == 5

    assert PartialApplication.lift(none).evaluate({})(1, 2) == 3

    assert PartialApplication.lift(all).evaluate({'A': 1, 'B': 2})() == 3

    @PartialApplication.lift
    def good_deco(a: float, b: float = Option('B')) -> float:
        return a + b

    assert good_deco.evaluate({'B': 2})(1) == 3

    @PartialApplication.lift(b=Option('B'))
    def good_deco_kwargs(a: float, b: float = Option('X')) -> float:
        return a + b

    assert good_deco_kwargs.evaluate({'B': 2})(1) == 3



import pytest

from labrea.application import FunctionApplication
from labrea import Value
from labrea.exceptions import KeyNotFoundError
from labrea.option import Option
from labrea.overload import Overloaded


def test_overloaded():
    @FunctionApplication.lift
    def add(x: int = Option('X'), y: int = Option('Y')) -> int:
        return x + y

    overloads = Overloaded(Option('A'), {}, default=add)

    @FunctionApplication.lift
    def dummy_add():
        return 0

    overloads.register('dummy', dummy_add)

    assert overloads.evaluate({'X': 1, 'Y': 2}) == 3
    assert overloads.evaluate({'A': 'dummy'}) == 0

    overloads.validate({'X': 1, 'Y': 2})
    overloads.validate({'A': 'dummy'})
    with pytest.raises(KeyNotFoundError):
        overloads.validate({'X': 1})

    assert overloads.keys({'X': 1, 'Y': 2}) == {'X', 'Y'}
    assert overloads.keys({'A': 'dummy'}) == {'A'}
    with pytest.raises(KeyNotFoundError):
        overloads.keys({'X': 1})

    assert overloads.explain({'X': 1, 'Y': 2}) == {'X', 'Y'}
    assert overloads.explain({'A': 'dummy'}) == {'A'}
    assert overloads.explain() == {'X', 'Y'}


def test_repr():
    assert repr(Overloaded(Option('A'), {'X': Value(42)})) == "Overloaded(Option('A'), {'X': Value(42)})"
    assert repr(Overloaded(Option('A'), {'X': Value(42)}, default=43)) == "Overloaded(Option('A'), {'X': Value(42)}, Value(43))"




from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types import Evaluatable


class EvaluationError(Exception):
    """Error raised when an object cannot be evaluated."""

    msg: str
    source: "Evaluatable"

    def __init__(self, msg: str, source: "Evaluatable") -> None:
        self.msg = msg
        self.source = source

        super().__init__(msg, source)

    def __str__(self):
        return f"Originating in {self.source} | {self.msg}"


class KeyNotFoundError(EvaluationError):
    """Error raised when a key is not found in the options dictionary."""

    key: str
    source: "Evaluatable"

    def __init__(self, key: str, source: "Evaluatable") -> None:
        self.key = key

        super().__init__(f"Key '{self.key}' not found", source)


class InsufficientInformationError(EvaluationError):
    """Error raised when not enough information is provided to explain an object."""

    def __init__(self, reason: str, source: "Evaluatable") -> None:
        super().__init__(
            f"Insufficient information to evaluate {source}: {reason}", source
        )



import pytest

from labrea.iterable import Iter, Map
from labrea.conditional import switch
from labrea.exceptions import EvaluationError, InsufficientInformationError
from labrea.option import Option


def test_iter():
    i = Iter(Option('A'), 2)

    assert i.apply(list).evaluate({'A': 1}) == [1, 2]
    i.validate({'A': 1})
    assert i.keys({'A': 1}) == {'A'}
    assert i.explain() == {'A'}

    with pytest.raises(EvaluationError):
        i.apply(list).evaluate({})
    with pytest.raises(EvaluationError):
        i.validate({})
    with pytest.raises(EvaluationError):
        i.keys({})

    assert repr(i) == "Iter(Option('A'), Value(2))"


def test_iterate():
    i = Map(Option('A'), {'A': Option('B')})

    assert i.apply(list).evaluate({'B': [1, 2]}) == [({'A': 1}, 1), ({'A': 2}, 2)]
    assert i.values.apply(list).evaluate({'B': [1, 2]}) == [1, 2]
    i.validate({'B': [1, 2]})
    assert i.keys({'B': [1, 2]}) == {'B'}
    assert i.explain({'B': [1, 2]}) == {'B'}

    with pytest.raises(EvaluationError):
        i.apply(list).evaluate({})
    with pytest.raises(EvaluationError):
        i.validate({})
    with pytest.raises(EvaluationError):
        i.keys({})
    assert i.explain() == {'B'}

    assert repr(i) == "Iterate(Option('A'), {'A': Option('B')})"


def test_iterate_explain_insufficient_information():
    i = Map(switch(Option('A'), {'X': 1}), {'A': Option('B')})

    with pytest.raises(InsufficientInformationError):
        i.explain()



from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    overload,
)

from confectioner import mix
from confectioner.templating import dotted_key_exists, get_dotted_key, resolve

from ._missing import MISSING, MaybeMissing
from .application import FunctionApplication
from .exceptions import KeyNotFoundError
from .template import Template
from .types import JSON, Evaluatable, MaybeEvaluatable, Options, Value

A = TypeVar("A", covariant=True, bound=JSON)
B = TypeVar("B", covariant=True)


class Option(Evaluatable[A]):
    """A class representing a single user-provided option.

    Options are the singular way to provide user-input to an Evaluatable
    object. They take a key, which is used to retrieve the value from the
    options dictionary. If the key does not exist, a default value can be
    provided.

    Arguments
    ----------
    key : str
        The key to retrieve from the options dictionary. This key can be
        nested using the standard :code:`{NESTED.KEY}` syntax from
        confectioner.
    default : MaybeEvaluatable[A]
        The default value to use if the key does not exist in the options
        dictionary. If the default value is an Evaluatable, it is evaluated
        using the options dictionary. If the default value is not an
        Evaluatable, it is returned as-is. If the default value is a string,
        it is treated as a :code:`Template` and evaluated using the options
        dictionary.
    default_factory : Callable[[], A]
        A factory function that returns the default value to use if the key
        does not exist in the options dictionary. This is an alternative to
        providing a default value directly.
    doc : str
        The docstring for the option.


    Example Usage
    -------------
    >>> from labrea import Option
    >>> o = Option('A.X', default='foo')
    >>> o()
    'foo'
    >>> o({'A': {'X': 'bar'}})
    'bar'
    """

    key: str
    default: MaybeMissing[Evaluatable[A]]

    def __init__(
        self,
        key: str,
        default: MaybeMissing[MaybeEvaluatable[A]] = MISSING,
        default_factory: MaybeMissing[Callable[[], A]] = MISSING,
        doc: str = "",
    ) -> None:
        self.key = key
        if isinstance(default, str):
            self.default = Template(default)  # type: ignore [assignment]
        elif default is not MISSING:
            self.default = Evaluatable.ensure(default)
        elif default_factory is not MISSING:
            self.default = FunctionApplication(default_factory)
        else:
            self.default = MISSING

        self.__doc__ = doc

    def evaluate(self, options: Options) -> A:
        """Retrieves the key from the options dictionary.

        If the key does not exist, the default key is returned. If no default
        key is provided, an error is raised. If the default key is an
        Evaluatable, it is evaluated using the options dictionary. If the
        default key is not an Evaluatable, it is returned as-is.
        """
        try:
            value = get_dotted_key(self.key, options)
            return resolve(value, options)
        except KeyError:
            if self.default is MISSING:
                raise KeyNotFoundError(self.key, self)
            return self.default.evaluate(options)

    def validate(self, options: Options) -> None:
        """Validates that the key exists in the options dictionary.

        If the key does not exist, the default key is validated. If the
        default key is an Evaluatable, it is validated using the options
        dictionary. If the default key is not an Evaluatable, it is ignored.
        """
        if dotted_key_exists(self.key, options):
            _ = self.keys(options)
        elif self.default is not MISSING:
            self.default.validate(options)
        else:
            raise KeyNotFoundError(self.key, self)

    def keys(self, options: Options) -> Set[str]:
        """Returns the keys required by the option.

        Usually this is just the provided key, but if the default value is a
        string, the keys required by the template are also returned. Similarly,
        if the default value is an Evaluatable, the keys required by the
        Evaluatable are also returned.
        """
        if dotted_key_exists(self.key, options):
            value = get_dotted_key(self.key, options)
            if isinstance(value, str):
                return {self.key} | Template(value).keys(options)
            else:
                return {self.key}
        elif self.default is not MISSING:
            return self.default.keys(options)
        else:
            raise KeyNotFoundError(self.key, self)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Returns the keys required by the option."""
        options = options or {}
        if dotted_key_exists(self.key, options):
            value = get_dotted_key(self.key, options)
            if isinstance(value, str):
                return {self.key} | Template(value).explain(options)
            else:
                return {self.key}
        elif self.default is not MISSING:
            return self.default.explain(options)
        else:
            return {self.key}

    def __repr__(self) -> str:
        return (
            f"Option({self.key!r}, default={self.default!r})"
            if self.default is not MISSING
            else f"Option({self.key!r})"
        )

    @overload
    @staticmethod
    def namespace(__namespace: Type) -> "Namespace": ...  # pragma: no cover

    @overload
    @staticmethod
    def namespace(
        __namespace: str,
    ) -> Callable[[Type], "Namespace"]: ...  # pragma: no cover

    @staticmethod
    def namespace(
        __namespace: Union[Type, str],
    ) -> Union["Namespace", Callable[[Type], "Namespace"]]:
        """Create an option namespace from a class definition

        This allows all of the options for a module to be grouped together. Namespaces can contain
        :class:`Option` objects or other namespaces.

        When developing a package with Labrea, it is recommended to create an option namespace for
        with the same name as the package to avoid conflicts with other packages.

        To create options with docstrings, see :meth:`Option.auto`.

        Example Usage
        -------------
        >>> from labrea import Option
        >>> @Option.namespace
        ... class MY_PACKAGE:
        ...     class MODULE_1:    # Implicit sub-namespace
        ...         A: str  # Equivalent to Option('MY_PACKAGE.MODULE_1.A')
        ...     @Option.namespace("MODULE-2")
        ...     class MODULE_2:    # Explicit sub-namespace with custom name
        ...         A = 10  # Equivalent to Option('MY_PACKAGE.MODULE_1.A', 10)
        ...
        >>> MY_PACKAGE.MODULE_2.A()
        10
        >>> MY_PACKAGE.MODULE_2.A({'MY_PACKAGE': {'MODULE-2': {'A': 20}}})
        20
        >>> print(MY_PACKAGE.__doc__)
        Namespace MY_PACKAGE:
            Namespace MY_PACKAGE.MODULE_1:
                Option MY_PACKAGE.MODULE_1.A
            Namespace MY_PACKAGE.MODULE-2:
                Option MY_PACKAGE.MODULE-2.A (default 10)
        """
        if isinstance(__namespace, str):
            return lambda cls: Namespace._from_type(cls, name=__namespace)
        return Namespace._from_type(__namespace)

    @staticmethod
    def auto(
        default: MaybeMissing[MaybeEvaluatable[A]] = MISSING,
        doc: str = "",
    ) -> "Option[A]":
        """Create an option in a namespace with an inferred key

        Sometimes when creating a namespace, we want to add an option with a docstring or some
        additional transformations, but we want the key to be inferred from the namespace. This
        function creates an option with a default value and docstring, but automatically infers
        the key.

        Example Usage
        -------------
        >>> from labrea import Option
        >>> @Option.namespace
        ... class MY_PACKAGE:
        ...     A = Option.auto(default='foo', doc='An automatic option') >> str
        ...
        >>> print(MY_PACKAGE.__doc__)
        Namespace MY_PACKAGE:
            Option MY_PACKAGE.A (default 'foo'): An automatic option
        >>> MY_PACKAGE.A()
        'foo'
        >>> MY_PACKAGE.A({'MY_PACKAGE': {'A': 100}})
        '100'  # str transformation applied
        """
        return _Auto(default, doc)  # type: ignore


class WithOptions(Evaluatable[B]):
    """A class that wraps an Evaluatable object and provides default options.

    This class is used to provide default options to an Evaluatable object. The
    default options are mixed with the provided options. The default options
    can be forced to take precedence over the provided options (default) or
    the provided options can take precedence over the default options.

    Arguments
    ----------
    evaluatable : Evaluatable[B]
        The evaluatable object to wrap.
    options : Options
        The default options to provide to the evaluatable object.
    force : bool, optional
        If True, the default options take precedence over the provided options.
        If False, the provided options take precedence over the default options.
        Default is True.
    """

    evaluatable: Evaluatable[B]
    options: Options
    force: bool

    def __init__(
        self, evaluatable: Evaluatable[B], options: Options, force: bool = True
    ) -> None:
        self.evaluatable = evaluatable
        self.options = options
        self.force = force

    def _options(self, options: Options) -> Options:
        return (
            mix(options, self.options)  # type: ignore
            if self.force
            else mix(self.options, options)  # type: ignore
        )

    def evaluate(self, options: Options) -> B:
        """Evaluate the wrapped Evaluatable object with the provided options."""
        return self.evaluatable.evaluate(self._options(options))

    def validate(self, options: Options) -> None:
        """Validate the wrapped Evaluatable object with the provided options."""
        self.evaluatable.validate(self._options(options))

    def keys(self, options: Options) -> Set[str]:
        """Return the keys required by the wrapped Evaluatable object."""
        return {
            key
            for key in self.evaluatable.keys(self._options(options))
            if not (
                dotted_key_exists(key, self.options)
                and (self.force or not dotted_key_exists(key, options))
            )
        }

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the explanation for the wrapped Evaluatable object."""
        options = options or {}
        return {
            key
            for key in self.evaluatable.explain(self._options(options))
            if not (
                dotted_key_exists(key, self.options)
                and (self.force or not dotted_key_exists(key, options))
            )
        }

    def __repr__(self) -> str:
        return (
            f"WithOptions({self.evaluatable!r}, {self.options!r})"
            if self.force
            else f"WithDefaultOptions({self.evaluatable!r}, {self.options!r})"
        )


def WithDefaultOptions(evaluatable: Evaluatable[B], options: Options) -> WithOptions[B]:
    """Wrap an Evaluatable object with default options.

    This is a convenience function for creating a WithOptions object
    with :code:`force=False`.

    Arguments
    ----------
    evaluatable : Evaluatable[B]
        The evaluatable object to wrap.
    options : Options
        The default options to provide to the evaluatable object.

    Returns
    -------
    WithOptions[B]
        The wrapped evaluatable object with default options.
    """
    return WithOptions(evaluatable, options, force=False)


class _AllOptions(Evaluatable[Options]):
    def evaluate(self, options: Options) -> Options:
        try:
            return resolve(options)
        except KeyError as e:
            raise KeyNotFoundError(e.args[0], self) from e

    def validate(self, options: Options) -> None:
        _ = self.evaluate(options)

    def keys(self, options: Options) -> Set[str]:
        return set(options.keys())

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return set() if options is None else set(options.keys())

    def __repr__(self) -> str:
        return "AllOptions"


AllOptions = _AllOptions()
AllOptions.__doc__ = """An object that evaluates to the entire options dictionary."""


class Namespace(Evaluatable[Options]):
    """Namespace for options that allows for better documentation and organization.

    This class should probably not be used directly. Instead, use the :meth:`Option.namespace`
    decorator.
    """

    _key: str
    _full: Option
    _members: Mapping[str, Union[Option, "_Auto", "Namespace"]]
    __doc__: str

    def __init__(
        self, key: str, members: Mapping[str, Union[Option, "_Auto", "Namespace"]]
    ) -> None:
        self._key = key
        self._full = Option(key, default={})
        self._members = members
        self.__doc__ = self._build_doc()

    def evaluate(self, options: Options) -> Options:
        return self._full.evaluate(options)

    def validate(self, options: Options) -> None:
        self._full.validate(options)

    def keys(self, options: Options) -> Set[str]:
        return self._full.keys(options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        return self._full.explain(options)

    def __getitem__(self, key: str) -> Evaluatable:
        item = self._members[key]
        if isinstance(item, _Auto):
            return item.build(f"{self._key}.{key}")
        return item

    def __getattr__(self, key: str) -> Evaluatable:
        try:
            return self[key]
        except KeyError:
            raise AttributeError(f"Namespace {self._key} has no attribute {key!r}")

    def __repr__(self) -> str:
        return f"Namespace({self._key!r})"

    @staticmethod
    def _build_doc_option(option: Option) -> str:
        doc = f"Option {option.key}"
        if option.default is not MISSING:
            default = (
                option.default.value
                if isinstance(option.default, Value)
                else option.default.template
                if isinstance(option.default, Template)
                else option.default
            )
            doc += f" (default {default!r})"
        if option.__doc__:
            doc += ": " + "\n  ".join(option.__doc__.strip().splitlines())
        return doc

    def _build_doc(self) -> str:
        members = sorted(self._members.items(), key=lambda x: x[0])
        options: List[Option] = []
        namespaces: List[Namespace] = []

        for key, value in members:
            if isinstance(value, Namespace):
                namespaces.append(value)
            elif isinstance(value, Option):
                options.append(value)
            elif isinstance(value, _Auto):
                options.append(value.option(f"{self._key}.{key}"))

        header = f"Namespace {self._key}:\n  "
        option_docs = (
            "\n  ".join(self._build_doc_option(opt).splitlines()) for opt in options
        )
        namespace_docs = (
            "\n  ".join((ns.__doc__ or "").splitlines()) + "\n" for ns in namespaces
        )

        return header + "\n  ".join((*option_docs, *namespace_docs)).rstrip()

    @classmethod
    def _from_type(
        cls, __namespace: Type, parent: Optional[str] = None, name: Optional[str] = None
    ) -> "Namespace":
        name = name or __namespace.__name__
        key = f"{parent}.{name}" if parent else name

        members: Dict[str, Union[Option, _Auto, Namespace]] = {}
        for name_ in getattr(__namespace, "__annotations__", {}):
            members[name_] = Option(f"{key}.{name_}")

        for name_, value in __namespace.__dict__.items():
            if isinstance(value, Namespace):
                members[name_] = value._inherit(key)
            elif isinstance(value, Option):
                if "." in value.key:
                    raise ValueError(
                        f"Namespace {key} has nested option {name_}: {value.key}"
                    )
                members[name_] = Option(
                    f"{key}.{value.key}", default=value.default, doc=value.__doc__ or ""
                )
            elif isinstance(value, _Auto):
                members[name_] = value
            elif name_.startswith("_"):
                continue
            elif isinstance(value, type):
                members[name_] = cls._from_type(value, parent=key)
            elif value is None or isinstance(
                value, (str, int, float, bool, list, dict, Evaluatable)
            ):
                members[name_] = Option(f"{key}.{name_}", default=value)
            else:
                raise TypeError(
                    f"Namespace {key} has non-JSON-serializable default value {name_}: {value!r}"
                )

        return cls(key, members)

    def _inherit(self, parent: str) -> "Namespace":
        return Namespace(
            f"{parent}.{self._key}",
            {
                name: (
                    value._inherit(parent)
                    if isinstance(value, Namespace)
                    else Option(
                        f"{parent}.{value.key}", value.default, doc=value.__doc__ or ""
                    )
                    if isinstance(value, Option)
                    else value
                )
                for name, value in self._members.items()
            },
        )


class _Auto(Generic[A]):
    default: MaybeMissing[MaybeEvaluatable[A]]
    doc: str
    transformations: List[Callable]

    def __init__(
        self,
        default: MaybeMissing[MaybeEvaluatable[A]] = MISSING,
        doc: str = "",
        *transformations: Callable,
    ) -> None:
        self.default = default
        self.doc = doc
        self.transformations = list(transformations)

    def option(self, key: str) -> Option[A]:
        return Option(key, self.default, doc=self.doc)

    def build(self, key: str, bare: bool = False) -> Evaluatable[A]:
        option: Evaluatable = self.option(key)

        for tform in self.transformations:
            option = option >> tform

        option.__doc__ = self.doc or option.__doc__

        return option

    def __rshift__(self, func: MaybeEvaluatable[Callable[[A], B]]) -> Evaluatable[B]:
        return _Auto(self.default, self.doc, *self.transformations, func)  # type: ignore



from typing import Any, List, Optional, Set, TypeVar

from .exceptions import EvaluationError
from .types import Evaluatable, Options

A = TypeVar("A", covariant=True)


class Coalesce(Evaluatable[A]):
    """Return the first Evaluatable that can be evaluated

    Takes 1 or more Evaluatables as arguments. Evaluates each Evaluatable in
    order until one can be evaluated. If none can be evaluated, raises an
    EvaluationError.

    Aliases: :func:`labrea.coalesce`, :class:`labrea.Coalesce`


    Arguments
    ---------
    *evaluatables : MaybeEvaluatable[A]
        The evaluatables to evaluate.


    Example Usage
    -------------
    >>> from labrea import Coalesce, Option
    >>> c = Coalesce(Option('A.X'), Option('B.Y'))
    >>> c({'A': {'X': 1}, 'B': {'Y': 2}})
    1
    >>> c({'B': {'Y': 2}})
    2
    """

    members: List[Evaluatable[A]]

    def __init__(self, __first: Evaluatable[A], *__rest: Evaluatable[A]):
        self.members = [Evaluatable.ensure(arg) for arg in (__first, *__rest)]

    def evaluate(self, options: Options) -> A:
        """Evaluate the first Evaluatable that can be evaluated"""
        return self._delegate("evaluate", options)

    def validate(self, options: Options) -> None:
        """Determines if any of the Evaluatables can be validated."""
        return self._delegate("validate", options)

    def keys(self, options: Options) -> Set[str]:
        """Return the keys for the first Evaluatable that can be evaluated."""
        return self._delegate("keys", options)

    def explain(self, options: Optional[Options] = None) -> Set[str]:
        """Return the explanation for the first Evaluatable that can be evaluated."""
        try:
            return self._delegate("explain", options)
        except EvaluationError:
            return self.members[-1].explain(options)

    def _delegate(self, method: str, options: Optional[Options] = None) -> Any:
        options = options or {}
        err: Optional[EvaluationError] = None

        for member in self.members:
            try:
                member.validate(options)
                return getattr(member, method)(options)
            except EvaluationError as e:
                err = e

        raise err  # type: ignore

    def __repr__(self) -> str:
        return f"Coalesce({', '.join(map(repr, self.members))})"


coalesce = Coalesce



import pytest
from typing import List

from labrea import pipeline_step, Option, Value
from labrea.exceptions import EvaluationError
from labrea.pipeline import Pipeline, PipelineStep


@pipeline_step
def incr(x: float) -> float:
    return x + 1


@pipeline_step
def repeat(x: float, n: int = Option('N')) -> List[float]:
    return [x for _ in range(n)]


incr_then_repeat = incr + repeat


def test_types():
    assert isinstance(incr, PipelineStep)
    assert isinstance(repeat, PipelineStep)
    assert isinstance(incr_then_repeat, Pipeline)


def test_bad_def():
    with pytest.raises(ValueError):
        @pipeline_step
        def missing_default(x: float, y: float) -> float:
            return x + y

    with pytest.raises(ValueError):
        @pipeline_step
        def unused_default(x: float = Option('X')):
            return x + 1


def test_pipeline_valid():
    assert incr_then_repeat.evaluate({'N': 3})(1) == [2, 2, 2]
    incr_then_repeat.validate({'N': 3})
    assert incr_then_repeat.keys({'N': 3}) == {'N'}
    assert incr_then_repeat.explain({'N': 3}) == {'N'}


def test_pipeline_invalid():
    with pytest.raises(EvaluationError):
        incr_then_repeat.evaluate({})
    with pytest.raises(EvaluationError):
        incr_then_repeat.validate({})
    with pytest.raises(EvaluationError):
        incr_then_repeat.keys({})
    assert incr_then_repeat.explain() == {'N'}


def test_repr():
    assert repr(incr) == '<PipelineStep incr>'
    assert repr(incr_then_repeat) == '<Pipeline <PipelineStep incr> + <PipelineStep repeat>>'
    assert repr(Pipeline()) == '<Pipeline Identity>'


def test_iter():
    assert list(incr_then_repeat) == [incr, repeat]


def test_add():
    def decr(x: float) -> float:
        return x - 1

    assert list(incr + repeat) == [incr, repeat]
    assert list(incr + repeat + decr) == [incr, repeat, PipelineStep(Value(decr))]
    assert list(incr_then_repeat + incr_then_repeat) == [incr, repeat, incr, repeat]
    assert list(Pipeline() + incr) == list(incr + Pipeline()) == [incr]


def test_transformation():
    assert incr.transform(1) == 2
    assert list(repeat.transform(1, {'N': 3})) == [1, 1, 1]
    assert list(incr_then_repeat.transform(1, {'N': 3})) == [2, 2, 2]
